from __future__ import annotations
import logging
from collections import defaultdict
import itertools
import pandas as pd
import numpy as np
import panel as pn
from IPython.display import display, Markdown

from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from bdikit.schema_matching.one2one.matcher_factory import SchemaMatchers
from bdikit.schema_matching.topk.base import BaseTopkSchemaMatcher
from bdikit.schema_matching.topk.matcher_factory import TopkMatchers
from bdikit.value_matching.base import BaseValueMatcher, ValueMatch, ValueMatchingResult
from bdikit.value_matching.matcher_factory import ValueMatchers
from bdikit.standards.standard_factory import Standards

from bdikit.mapping_functions import (
    ValueMapper,
    FunctionValueMapper,
    DictionaryMapper,
    IdentityValueMapper,
)

from typing import (
    Union,
    List,
    Dict,
    TypedDict,
    Optional,
    Tuple,
    Callable,
    Any,
)

from bdikit.config import DEFAULT_SCHEMA_MATCHING_METHOD, DEFAULT_VALUE_MATCHING_METHOD

pn.extension("tabulator")

logger = logging.getLogger(__name__)


def match_schema(
    source: pd.DataFrame,
    target: Union[str, pd.DataFrame] = "gdc",
    method: Union[str, BaseSchemaMatcher] = DEFAULT_SCHEMA_MATCHING_METHOD,
    method_args: Optional[Dict[str, Any]] = None,
) -> pd.DataFrame:
    """
    Performs schema mapping between the source table and the given target schema. The
    target is either a DataFrame or a string representing a standard data vocabulary
    supported by the library. Currently, only the GDC (Genomic Data Commons) standard
    vocabulary is supported.

    Parameters:
        source (pd.DataFrame): The source table to be mapped.
        target (Union[str, pd.DataFrame], optional): The target table or standard data vocabulary. Defaults to "gdc".
        method (str, optional): The method used for mapping. Defaults to "ct_learning".
        method_args (Dict[str, Any], optional): The additional arguments of the method for schema matching.

    Returns:
        pd.DataFrame: A DataFrame containing the mapping results with columns "source" and "target".

    Raises:
        ValueError: If the method is neither a string nor an instance of BaseColumnMappingAlgorithm.
    """
    if isinstance(target, str):
        target_table = _load_table_for_standard(target)
    else:
        target_table = target

    if isinstance(method, str):
        if method_args is None:
            method_args = {}
        matcher_instance = SchemaMatchers.get_matcher(method, **method_args)
    elif isinstance(method, BaseSchemaMatcher):
        matcher_instance = method
    else:
        raise ValueError(
            "The method must be a string or an instance of BaseColumnMappingAlgorithm"
        )

    matches = matcher_instance.map(source, target_table)

    return pd.DataFrame(matches.items(), columns=["source", "target"])


def _load_table_for_standard(name: str) -> pd.DataFrame:
    """
    Load the table for the given standard data vocabulary. Currently, only the
    GDC standard is supported.
    """
    standard = Standards.get_standard(name)
    df = standard.get_dataframe_rep()

    return df


def top_matches(
    source: pd.DataFrame,
    columns: Optional[List[str]] = None,
    target: Union[str, pd.DataFrame] = "gdc",
    top_k: int = 10,
    method: Union[str, BaseTopkSchemaMatcher] = DEFAULT_SCHEMA_MATCHING_METHOD,
    method_args: Optional[Dict[str, Any]] = None,
) -> pd.DataFrame:
    """
    Returns the top-k matches between the source and target tables.

    Args:
        source (pd.DataFrame): The source table.
        columns (Optional[List[str]], optional): The list of columns to consider for matching. Defaults to None.
        target (Union[str, pd.DataFrame], optional): The target table or the name of the standard target table. Defaults to "gdc".
        top_k (int, optional): The number of top matches to return. Defaults to 10.

    Returns:
        pd.DataFrame: A DataFrame containing the top-k matches between the source and target tables.
    """

    if isinstance(target, str):
        target_table = _load_table_for_standard(target)
    else:
        target_table = target

    if columns is not None and len(columns) > 0:
        selected_columns = source[columns]
    else:
        selected_columns = source

    if isinstance(method, str):
        if method_args is None:
            method_args = {}
        topk_matcher = TopkMatchers.get_matcher(method, **method_args)
    elif isinstance(method, BaseTopkSchemaMatcher):
        topk_matcher = method
    else:
        raise ValueError(
            "The method must be a string or an instance of BaseTopkColumnMatcher"
        )

    top_k_matches = topk_matcher.get_recommendations(
        selected_columns, target=target_table, top_k=top_k
    )

    dfs = []
    for match in top_k_matches:
        matches = pd.DataFrame(match["top_k_columns"], columns=["target", "similarity"])
        matches["source"] = match["source_column"]
        matches = matches[["source", "target", "similarity"]]  # reorder columns
        dfs.append(matches.sort_values(by="similarity", ascending=False))

    return pd.concat(dfs, ignore_index=True)


def match_values(
    source: pd.DataFrame,
    target: Union[str, pd.DataFrame],
    column_mapping: Union[Tuple[str, str], pd.DataFrame],
    method: Union[str, BaseValueMatcher] = DEFAULT_VALUE_MATCHING_METHOD,
    method_args: Optional[Dict[str, Any]] = None,
) -> Union[pd.DataFrame, List[pd.DataFrame]]:
    """
    Finds matches between column values from the source dataset and column
    values of the target domain (a pd.DataFrame or a standard dictionary such
    as 'gdc') using the method provided in `method`.

    Args:
        source (pd.DataFrame): The source dataset containing the columns to be
          matched.

        target (Union[str, pd.DataFrame]): The target domain to match the
          values to. It can be either a DataFrame or a standard vocabulary name.

        column_mapping (Union[Tuple[str, str], pd.DataFrame]): A tuple or a
          DataFrame containing the mappings between source and target columns.

          - If a tuple is provided, it should contain two strings where the first
            is the source column and the second is the target column.
          - If a DataFrame is provided, it should contain 'source' and 'target'
            column names where each row specifies a column mapping.

        method (str, optional): The name of the method to use for value
          matching.
        method_args (Dict[str, Any], optional): The additional arguments of the
            method for value matching.

    Returns:
        Union[pd.DataFrame, List[pd.DataFrame]]: A list of DataFrame objects containing
        the results of value matching between the source and target values. If a tuple
        is provided as the `column_mapping`, only a DataFrame instance is returned.

    Raises:
        ValueError: If the column_mapping DataFrame does not contain 'source' and
          'target' columns.
        ValueError: If the target is neither a DataFrame nor a standard vocabulary name.
        ValueError: If the source column is not present in the source dataset.
    """
    if method_args is None:
        method_args = {}

    if "top_k" in method_args and method_args["top_k"] > 1:
        logger.warning(
            f"Ignoring 'top_k' argument, use the 'top_value_matches()' method to get top-k value matches."
        )
        method_args["top_k"] = 1

    matches = _match_values(source, target, column_mapping, method, method_args)

    if isinstance(column_mapping, tuple):
        if len(matches) == 0:
            return pd.DataFrame(columns=["source", "target", "similarity"])
        # If only a single mapping is provided (as a tuple), we return the result
        # directly as a DataFrame to make it easier to display it in notebooks.
        assert (
            len(matches) == 1
        ), f"Expected one result for a single column mapping, but got: {len(matches)}"
        return matches[0]
    else:
        return matches


def top_value_matches(
    source: pd.DataFrame,
    target: Union[str, pd.DataFrame],
    column_mapping: Union[Tuple[str, str], pd.DataFrame],
    top_k: int = 5,
    method: str = DEFAULT_VALUE_MATCHING_METHOD,
    method_args: Optional[Dict[str, Any]] = None,
) -> List[pd.DataFrame]:
    """
    Finds top value matches between column values from the source dataset and column
    values of the target domain (a pd.DataFrame or a standard dictionary such
    as 'gdc') using the method provided in `method`.

    Args:
        source (pd.DataFrame): The source dataset containing the columns to be
          matched.

        target (Union[str, pd.DataFrame]): The target domain to match the
          values to. It can be either a DataFrame or a standard vocabulary name.

        column_mapping (Union[Tuple[str, str], pd.DataFrame]): A tuple or a
          DataFrame containing the mappings between source and target columns.

          - If a tuple is provided, it should contain two strings where the first
            is the source column and the second is the target column.
          - If a DataFrame is provided, it should contain 'source' and 'target'
            column names where each row specifies a column mapping.

        top_k (int, optional): The number of top matches to return. Defaults to 5.

        method (str, optional): The name of the method to use for value
          matching.
        method_args (Dict[str, Any], optional): The additional arguments of the
            method for value matching.

    Returns:
        List[pd.DataFrame]: A list of DataFrame objects containing
        the results of value matching between the source and target values.

    Raises:
        ValueError: If the column_mapping DataFrame does not contain 'source' and
          'target' columns.
        ValueError: If the target is neither a DataFrame nor a standard vocabulary name.
        ValueError: If the source column is not present in the source dataset.
    """
    if method_args is None:
        method_args = {}

    if "top_k" in method_args:
        logger.warning(
            f"Ignoring 'top_k' argument, using top_k argument instead (top_k={top_k})"
        )

    method_args["top_k"] = top_k

    matches = _match_values(source, target, column_mapping, method, method_args)

    match_list = []
    for match in matches:
        for _, group in match.groupby("source", dropna=False):
            match_list.append(
                group.reset_index(drop=True).sort_values(
                    by=["similarity"], ascending=False
                )
            )

    return match_list


def view_value_matches(
    matches: Union[pd.DataFrame, List[pd.DataFrame]], edit: bool = False
):
    """
    Shows the value match results in a DataFrame fashion.

    Args:
        matches (Union[pd.DataFrame, List[pd.DataFrame]]): The value match results
          obtained by the method match_values().

        edit (bool): Whether or not to edit the values within the DataFrame.
    """
    if isinstance(matches, pd.DataFrame):
        match_list = [matches]
    elif isinstance(matches, list):
        match_list = matches
    else:
        raise ValueError("The matches must be a DataFrame or a list of DataFrames")

    # Grouping DataFrames by metadata (source and target columns)
    grouped_matches = defaultdict(list)
    for match_df in match_list:
        grouped_matches[match_df.attrs["source"], match_df.attrs["target"]].append(
            match_df
        )

    # Display grouped DataFrames
    for (source_col, target_col), match_dfs in grouped_matches.items():
        display(
            Markdown(
                f"<br>**Source column:** {source_col}<br>"
                f"**Target column:** {target_col}<br>"
            )
        )
        for match_df in match_dfs:
            if edit:
                match_widget = pn.widgets.Tabulator(match_df, disabled=not edit)
                display(match_widget)
            else:
                display(match_df)


def _match_values(
    source: pd.DataFrame,
    target: Union[str, pd.DataFrame],
    column_mapping: Union[Tuple[str, str], pd.DataFrame],
    method: str,
    method_args: Dict[str, Any],
) -> List[pd.DataFrame]:

    target_domain, column_mapping_list = _format_value_matching_input(
        source, target, column_mapping
    )
    value_matcher = ValueMatchers.get_matcher(method, **method_args)
    mapping_results: List[ValueMatchingResult] = []

    for mapping in column_mapping_list:
        source_column, target_column = mapping["source"], mapping["target"]

        # 1. Select candidate columns for value mapping
        target_domain_list = target_domain[target_column]
        if target_domain_list is None or len(target_domain_list) == 0:
            continue

        unique_values = source[source_column].unique()
        if _skip_values(unique_values):
            continue

        # 2. Remove blank spaces to the unique values
        source_values_dict: Dict[str, Any] = {str(x).strip(): x for x in unique_values}
        target_values_dict: Dict[str, str] = {
            str(x).strip(): x for x in target_domain_list
        }

        # 3. Apply the value matcher to create value mapping dictionaries
        raw_matches = value_matcher.match(
            list(source_values_dict.keys()), list(target_values_dict.keys())
        )

        # 4. Transform the matches to the original
        matches: List[ValueMatch] = []
        for source_value, target_value, similarity in raw_matches:
            matches.append(
                ValueMatch(
                    source_value=source_values_dict[source_value],
                    target_value=target_values_dict[target_value],
                    similarity=similarity,
                )
            )

        # 5. Calculate the coverage and unmatched values
        source_values = set(source_values_dict.values())
        match_values = set([x[0] for x in matches])
        coverage = len(match_values) / len(source_values_dict)

        mapping_results.append(
            ValueMatchingResult(
                source=source_column,
                target=target_column,
                matches=matches,
                coverage=coverage,
                unique_values=source_values,
                unmatch_values=source_values - match_values,
            )
        )

    mapping_df_list = [
        _value_matching_result_to_df(mapping_result)
        for mapping_result in mapping_results
    ]

    return mapping_df_list


def _format_value_matching_input(
    source: pd.DataFrame,
    target: Union[str, pd.DataFrame],
    column_mapping: Union[Tuple[str, str], pd.DataFrame],
):
    if isinstance(column_mapping, pd.DataFrame):
        if not all(k in column_mapping.columns for k in ["source", "target"]):
            raise ValueError(
                "The column_mapping DataFrame must contain 'source' and 'target' columns."
            )
        mapping_df = column_mapping
    elif isinstance(column_mapping, tuple):
        mapping_df = pd.DataFrame(
            [
                {
                    "source": column_mapping[0],
                    "target": column_mapping[1],
                }
            ]
        )
    else:
        raise ValueError(
            "The column_mapping must be a DataFrame or a tuple of two strings "
            "containing the 'source' and 'target' columns."
        )

    column_mapping_list = mapping_df.to_dict(orient="records")

    for mapping in column_mapping_list:
        source_column = mapping["source"]
        if source_column not in source.columns:
            raise ValueError(
                f"The source column '{source_column}' is not present in the source dataset."
            )

    if isinstance(target, str):
        column_names = mapping_df["target"].unique().tolist()
        standard = Standards.get_standard(target)
        target_domain = standard.get_column_values(column_names)
    elif isinstance(target, pd.DataFrame):
        target_domain = {
            column_name: target[column_name].unique().tolist()
            for column_name in target.columns
        }
    else:
        raise ValueError(
            "The target must be a DataFrame or a standard vocabulary name."
        )

    return target_domain, column_mapping_list


def _value_matching_result_to_df(
    matching_result: ValueMatchingResult, default_unmatched: Any = np.nan
) -> pd.DataFrame:
    """
    Transforms the list of matches and unmatched values into a DataFrame.
    """
    matches_df = pd.DataFrame(
        data=matching_result["matches"],
        columns=["source", "target", "similarity"],
    )

    unmatched_values = matching_result["unmatch_values"]

    unmatched_df = pd.DataFrame(
        data=list(
            zip(
                unmatched_values,
                [default_unmatched] * len(unmatched_values),
                [default_unmatched] * len(unmatched_values),
            )
        ),
        columns=["source", "target", "similarity"],
    )

    result = pd.concat([matches_df, unmatched_df], ignore_index=True)
    result.attrs["source"] = matching_result["source"]
    result.attrs["target"] = matching_result["target"]
    result.attrs["coverage"] = matching_result["coverage"]
    return result


def _skip_values(unique_values: np.ndarray):
    if pd.api.types.is_numeric_dtype(unique_values):
        return True
    else:
        return False


def preview_domain(
    dataset: Union[str, pd.DataFrame],
    column: str,
    limit: Optional[int] = None,
) -> pd.DataFrame:
    """
    Preview the domain, i.e. set of unique values, column description and value description
    (if applicable) of the given column of the source or target dataset.

    Args:
        dataset (Union[str, pd.DataFrame], optional): The dataset or standard vocabulary name
        containing the column to preview.
            If a string is provided and it is equal to "gdc", the domain will be retrieved
            from the GDC data.
            If a DataFrame is provided, the domain will be retrieved from the specified DataFrame.
        column(str): The column name to show the domain.
        limit (int, optional): The maximum number of unique values to include in the preview.
            Defaults to None.

    Returns:
        pd.DataFrame: A DataFrame containing the unique domain values (or a sample of
        them if the parameter `limit` was specified), column description and value description
        (if applicable).
    """

    if isinstance(dataset, str):
        standard = Standards.get_standard(dataset)
        column_metadata = standard.get_column_metadata([column])
        value_names = column_metadata[column]["value_names"]
        value_descriptions = column_metadata[column]["value_descriptions"]
        column_description = column_metadata[column]["description"]
        assert len(value_names) == len(value_descriptions)
    elif isinstance(dataset, pd.DataFrame):
        value_names = dataset[column].unique()
        value_descriptions = []
        column_description = ""
    else:
        raise ValueError(
            "The dataset must be a DataFrame or a standard vocabulary name."
        )

    if isinstance(limit, int):
        value_names = value_names[:limit]
        value_descriptions = value_descriptions[:limit]

    domain = {}

    if len(value_names) > 0:
        domain["value_name"] = value_names

    if len(value_descriptions) > 0:
        domain["value_description"] = value_descriptions

    if len(column_description) > 0:
        empty_rows_size = len(value_names) - 1
        domain["column_description"] = [column_description] + [""] * empty_rows_size

    return pd.DataFrame(domain)


class ColumnMappingSpec(TypedDict):
    source: str
    target: str
    mapper: ValueMapper


def merge_mappings(
    mappings: MappingSpecLike, user_mappings: Optional[MappingSpecLike] = None
) -> List:
    """
    Creates a "data harmonization" plan based on the provided schema and/or value mappings.
    These mappings can either be computed by the library's functions or provided by the user.
    If the user mappings are provided (using the user_mappings parameter), they will take
    precedence over the mappings provided in the first parameter.

    Args:
        mappings (MappingSpecLike): The value mappings used to create the data
            harmonization plan. It can be a DataFrame, a list of dictionaries or a list of DataFrames.
        user_mappings (Optional[MappingSpecLike]): The user mappings to be included in
            the update. It can be a DataFrame, a list of dictionaries or a list of DataFrames.
            Defaults to None.

    Returns:
        List: The data harmonization plan that can be used as input to the :py:func:`~bdikit.api.materialize_mapping()`
        function. Concretely, the harmonization plan is a list of dictionaries, where each
        dictionary contains the source column, target column, and mapper object that will be used
        to transform the input to the output data.

    Raises:
        ValueError: If there are duplicate mappings for the same source and target columns.

    """
    if user_mappings is None:
        user_mappings = []

    mapping_spec_list = _normalize_mapping_spec(mappings)
    user_mapping_spec_list = _normalize_mapping_spec(user_mappings)

    def create_key(source: str, target: str) -> str:
        return source + "__" + target

    def check_duplicates(mappings: List[ColumnMappingSpec]):
        keys = set()
        for mapping in mappings:
            key = create_key(mapping["source"], mapping["target"])
            if key in keys:
                raise ValueError(
                    f"Duplicate mapping for source: {mapping['source']}, target: {mapping['target']}"
                )
            keys.add(key)

    # first check duplicates in each individual list
    check_duplicates(user_mapping_spec_list)
    check_duplicates(mapping_spec_list)

    mapping_keys = set()
    final_mappings = []

    # include all unique user mappings first, as they take precedence
    for mapping in itertools.chain(user_mapping_spec_list, mapping_spec_list):

        source_column = mapping["source"]
        target_column = mapping["target"]

        # ignore duplicate mappings across user and value mappings
        key = create_key(source_column, target_column)
        if key in mapping_keys:
            continue
        else:
            mapping_keys.add(key)

        # try creating a mapper object from the mapping
        mapper = create_mapper(mapping)

        final_mappings.append(
            {
                "source": source_column,
                "target": target_column,
                "mapper": mapper,
            }
        )

    return final_mappings


def _normalize_mapping_spec(mapping_spec: MappingSpecLike) -> List[ColumnMappingSpec]:

    if isinstance(mapping_spec, pd.DataFrame):
        mapping_spec_list: List = mapping_spec.to_dict(orient="records")
    elif isinstance(mapping_spec, List):
        mapping_spec_list: List = mapping_spec

    normalized: List[ColumnMappingSpec] = []
    for mapping_spec in mapping_spec_list:
        if isinstance(mapping_spec, pd.DataFrame):
            mapping_dict = _df_to_mapping_spec_dict(mapping_spec)
        elif isinstance(mapping_spec, Dict):
            mapping_dict = mapping_spec
        else:
            raise ValueError(
                f"Each mapping specification must be a dictionary or a DataFrame,"
                f" but was: {str(mapping_spec)}"
            )

        if "source" not in mapping_dict or "target" not in mapping_dict:
            raise ValueError(
                "Each mapping specification should contain 'source', 'target' "
                f"and 'mapper' (optional) keys but found only {mapping_dict.keys()}."
            )

        if "mapper" in mapping_dict and isinstance(mapping_dict["mapper"], ValueMapper):
            mapper = mapping_dict["mapper"]
        else:
            mapper = create_mapper(mapping_dict)

        normalized.append(
            {
                "source": mapping_dict["source"],
                "target": mapping_dict["target"],
                "mapper": mapper,
            }
        )

    return normalized


def _df_to_mapping_spec_dict(spec: Union[Dict, pd.DataFrame]) -> Dict:
    if isinstance(spec, Dict):
        return spec
    elif isinstance(spec, pd.DataFrame):
        if "source" not in spec.attrs or "target" not in spec.attrs:
            raise ValueError(
                "The DataFrame must contain 'source' and 'target' attributes."
            )
        return {
            "source": spec.attrs["source"],
            "target": spec.attrs["target"],
            "matches": spec,
        }
    else:
        raise ValueError(f"Invalid mapping specification: {str(spec)}")


def materialize_mapping(
    input_table: pd.DataFrame, mapping_spec: MappingSpecLike
) -> pd.DataFrame:
    """
    Takes an input DataFrame and a target mapping specification and returns a
    new DataFrame created according to the given target mapping specification.
    The mapping specification is a list of dictionaries, where each dictionary
    defines one column in the output table and how it is created. It includes
    the names of the input (source) and output (target) columns and the value
    mapper used to transform the values of the input column into the
    target output column.

    Parameters:
        input_table (pd.DataFrame): The input (source) DataFrame.
        mapping_spec (MappingSpecLike): The target mapping specification. It can
            be a DataFrame, a list of dictionaries or a list of DataFrames.

    Returns:
        pd.DataFrame: A DataFrame, which is created according to the target
        mapping specifications.
    """
    mapping_spec_list = _normalize_mapping_spec(mapping_spec)

    for mapping in mapping_spec_list:
        if mapping["source"] not in input_table.columns:
            raise ValueError(
                f"The source column '{mapping['source']}' is not present in "
                " the input table."
            )

    # execute the actual mapping plan
    output_dataframe = pd.DataFrame()
    for column_spec in mapping_spec_list:
        from_column_name = column_spec["source"]
        to_column_name = column_spec["target"]
        value_mapper = column_spec["mapper"]
        output_dataframe[to_column_name] = value_mapper.map(
            input_table[from_column_name]
        )
    return output_dataframe


def create_mapper(
    input: Union[
        None,
        ValueMapper,
        pd.DataFrame,
        ValueMatchingResult,
        List[ValueMatch],
        Dict,
        ColumnMappingSpec,
        Callable[[pd.Series], pd.Series],
    ]
):
    """
    Tries to instantiate an appropriate ValueMapper object for the given input argument.
    Depending on the input type, it may create one of the following objects:

    - If input is None, it creates an IdentityValueMapper object.
    - If input is a ValueMapper, it returns the input object.
    - If input is a function (or lambda function), it creates a FunctionValueMapper object.
    - If input is a list of ValueMatch objects or tuples (<source_value>, <target_value>),
      it creates a DictionaryMapper object.
    - If input is a DataFrame with two columns ("source_value", "target_value"),
      it creates a DictionaryMapper object.
    - If input is a dictionary containing a "source" and "target" key, it tries to create
      a ValueMapper object based on the specification given in "mapper" or "matches" keys.

    Args:
        input: The input argument to create a ValueMapper object from.

    Returns:
        ValueMapper: An instance of a ValueMapper.
    """
    # If no input is provided, we create an IdentityValueMapper by default
    # to not change the values from the source column
    if input is None:
        return IdentityValueMapper()

    # If the input is already a ValueMapper, no need to create a new one
    if isinstance(input, ValueMapper):
        return input

    # If the input is a function, we can create a FunctionValueMapper
    # that applies the function to the values of the source column
    if callable(input):
        return FunctionValueMapper(input)

    # This could be a list of value matches produced by match_values(),
    # so can create a DictionaryMapper based on the value matches
    if isinstance(input, List):
        return _create_mapper_from_value_matches(input)

    # If the input is a DataFrame with two columns, we can create a
    # DictionaryMapper based on the values in the DataFrame
    if isinstance(input, pd.DataFrame) and all(
        k in input.columns for k in ["source_value", "target_value"]
    ):
        return DictionaryMapper(
            input.set_index("source_value")["target_value"].to_dict()
        )

    if isinstance(input, Dict):
        if all(k in input for k in ["source", "target"]):
            # This could be the mapper created by merge_mappings() or a
            # specification defined by the user
            if "mapper" in input:
                if isinstance(input["mapper"], ValueMapper):
                    # If it contains a ValueMapper object, just return it
                    return input["mapper"]
                else:
                    # Else, 'mapper' may contain one of the basic values that
                    # can be used to create a ValueMapper object defined above,
                    # so call this function recursively create it
                    return create_mapper(input["mapper"])

            # This could be the a list of value matches (i.e., ValueMatch
            # or tuple(source, target)) provided by the user
            if "matches" in input and isinstance(input["matches"], List):
                return _create_mapper_from_value_matches(input["matches"])

            if "matches" in input and isinstance(input["matches"], pd.DataFrame):
                # This could be the output of match_values(), so we can
                # create a DictionaryMapper based on the value matches
                return DictionaryMapper(
                    input["matches"].set_index("source")["target"].to_dict()
                )

            # This could be the output of match_schema(), but the user did not
            # define any mapper, so we create an IdentityValueMapper to map the
            # column to the target name but keeping the values as they are
            return IdentityValueMapper()

    raise ValueError(f"Failed to create a ValueMapper for given input: {input}")


def _create_mapper_from_value_matches(matches: List[ValueMatch]) -> DictionaryMapper:
    mapping_dict = {}
    for match in matches:
        if isinstance(match, ValueMatch):
            mapping_dict[match.source_value] = match.target_value
        elif isinstance(match, tuple) and len(match) == 2:
            if isinstance(match[0], str) and isinstance(match[1], str):
                mapping_dict[match[0]] = match[1]
            else:
                raise ValueError(
                    "Tuple in matches must contain two strings: (source_value, target_value)"
                )
        else:
            raise ValueError("Matches must be a list of ValueMatch objects or tuples")
    return DictionaryMapper(mapping_dict)


MappingSpecLike = Union[List[Union[Dict, pd.DataFrame]], pd.DataFrame]
"""
The `MappingSpecLike` is a type alias that specifies mappings between source
and target columns. It must include the source and target column names
and a value mapper object that transforms the values of the source column
into the target.

The mapping specification can be (1) a DataFrame or (2) a list of dictionaries or DataFrames.

If it is a list of dictionaries, they must have:

- `source`: The name of the source column.
- `target`: The name of the target column.
- `mapper` (optional): A ValueMapper instance or an object that can be used to
  create one using :py:func:`~bdikit.api.create_mapper()`. Examples of valid objects
  are Python functions or lambda functions. If empty, an IdentityValueMapper
  is used by default.
- `matches` (optional): Specifies the value mappings. It can be a DataFrame containing
  the matches (returned by :py:func:`~bdikit.api.match_values()`), a list of ValueMatch
  objects, or a list of tuples (<source_value>, <target_value>).

Alternatively, the list can contain DataFrames. In this case, the DataFrames must
contain not only the value mappings (as described in the `matches` key above) but
also the `source` and `target` columns as DataFrame attributes. The DataFrames created
by :py:func:`~bdikit.api.match_values()` include this information by default.

If the mapping specification is a DataFrame, it must be compatible with the dictionaries
above and contain `source`, `target`, and `mapper` or `matcher` columns.

Example:

.. code-block:: python

    mapping_spec = [
      {
        "source": "source_column1",
        "target": "target_column1",
      },
      {
        "source": "source_column2",
        "target": "target_column2",
        "mapper": lambda age: -age * 365.25,
      },
      {
        "source": "source_column3",
        "target": "target_column3",
        "matches": [
          ("source_value1", "target_value1"),
          ("source_value2", "target_value2"),
        ]
      },
      {
        "source": "source_column",
        "target": "target_column",
        "matches": df_value_mapping_1
      },
      df_value_mapping_2, # a DataFrame returned by match_values()
    ]
"""



import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers import AutoModel, AutoTokenizer, logging

# Suppress the warning messages about untrained layer when loading the roberta  model
# in line: self.bert = AutoModel.from_pretrained(lm_mp[lm])
logging.set_verbosity_error()

lm_mp = {"roberta": "roberta-base", "distilbert": "distilbert-base-uncased"}


class BarlowTwinsSimCLR(nn.Module):
    """Barlow Twins or SimCLR encoder for contrastive learning."""

    def __init__(self, scale_loss, lambd, projector=768, device="cuda", lm="roberta"):
        super().__init__()
        self.bert = AutoModel.from_pretrained(lm_mp[lm])
        self.device = device
        self.scale_loss = scale_loss
        self.lambd = lambd
        hidden_size = 768

        # projector
        self.projector = nn.Linear(hidden_size, projector)

        # normalization layer for the representations z1 and z2
        self.bn = nn.BatchNorm1d(hidden_size, affine=False)

        # a fully connected layer for fine tuning
        self.fc = nn.Linear(hidden_size * 2, 2)

        # contrastive
        self.criterion = nn.CrossEntropyLoss().to(device)

        # cls token id
        self.cls_token_id = AutoTokenizer.from_pretrained(lm_mp[lm]).cls_token_id

    def info_nce_loss(self, features, batch_size, n_views, temperature=0.07):
        """Copied from https://github.com/sthalles/SimCLR/blob/master/simclr.py"""
        labels = torch.cat([torch.arange(batch_size) for i in range(n_views)], dim=0)
        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        labels = labels.to(self.device)

        features = F.normalize(features, dim=1)

        similarity_matrix = torch.matmul(features, features.T)

        # discard the main diagonal from both: labels and similarities matrix
        mask = torch.eye(labels.shape[0], dtype=torch.bool).to(self.device)
        labels = labels[~mask].view(labels.shape[0], -1)
        similarity_matrix = similarity_matrix[~mask].view(
            similarity_matrix.shape[0], -1
        )
        # assert similarity_matrix.shape == labels.shape

        # select and combine multiple positives
        positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

        # select only the negatives the negatives
        negatives = similarity_matrix[~labels.bool()].view(
            similarity_matrix.shape[0], -1
        )

        logits = torch.cat([positives, negatives], dim=1)
        labels = torch.zeros(logits.shape[0], dtype=torch.long).to(self.device)

        logits = logits / temperature
        return logits, labels

    def _extract_columns(self, x, z, cls_indices=None):
        """Helper function for extracting column vectors from LM outputs."""
        x_flat = x.view(-1)
        column_vectors = z.view((x_flat.shape[0], -1))

        if cls_indices is None:
            indices = [
                idx
                for idx, token_id in enumerate(x_flat)
                if token_id == self.cls_token_id
            ]
        else:
            indices = []
            seq_len = x.shape[-1]
            for rid in range(len(cls_indices)):
                indices += [idx + rid * seq_len for idx in cls_indices[rid]]

        return column_vectors[indices]

    def inference(self, x):
        """Apply the model on a serialized table.

        Args:
            x (LongTensor): a batch of serialized tables

        Returns:
            Tensor: the column vectors for all tables
        """
        x = x.to(self.device)
        z = self.bert(x)[0]
        z = self.projector(z)  # optional
        return self._extract_columns(x, z)

    def forward(self, x_ori, x_aug, cls_indices, mode="simclr"):
        """Apply the model for contrastive learning.

        Args:
            x_ori (LongTensor): the first views of a batch of tables
            x_aug (LongTensor): the second views of a batch of tables
            cls_indices (tuple of List): the cls_token alignment
            mode (str, optional): the name of the contrastive learning algorithm

        Returns:
            Tensor: the loss
        """
        if mode in ["simclr", "barlow_twins"]:
            # pre-training
            # encode
            batch_size = len(x_ori)
            x_ori = x_ori.to(self.device)  # original, (batch_size, seq_len)
            x_aug = x_aug.to(self.device)  # augment, (batch_size, seq_len)

            # encode each table (all columns)
            x = torch.cat((x_ori, x_aug))  # (2*batch_size, seq_len)
            z = self.bert(x)[0]  # (2*batch_size, seq_len, hidden_size)

            # assert that x_ori and x_aug have the same number of columns
            z_ori = z[:batch_size]  # (batch_size, seq_len, hidden_size)
            z_aug = z[batch_size:]  # (batch_size, seq_len, hidden_size)

            cls_ori, cls_aug = cls_indices

            z_ori = self._extract_columns(
                x_ori, z_ori, cls_ori
            )  # (total_num_columns, hidden_size)
            z_aug = self._extract_columns(
                x_aug, z_aug, cls_aug
            )  # (total_num_columns, hidden_size)
            assert z_ori.shape == z_aug.shape

            z = torch.cat((z_ori, z_aug))
            z = self.projector(z)  # (2*total_num_columns, projector_size)

            if mode == "simclr":
                # simclr
                logits, labels = self.info_nce_loss(z, len(z) // 2, 2)
                loss = self.criterion(logits, labels)
                return loss
            elif mode == "barlow_twins":
                # barlow twins
                z1 = z[: len(z) // 2]
                z2 = z[len(z) // 2 :]

                # empirical cross-correlation matrix
                c = (self.bn(z1).T @ self.bn(z2)) / (len(z1))

                # use --scale-loss to multiply the loss by a constant factor
                on_diag = ((torch.diagonal(c) - 1) ** 2).sum() * self.scale_loss
                off_diag = (off_diagonal(c) ** 2).sum() * self.scale_loss
                loss = on_diag + self.hp.lambd * off_diag
                return loss
        elif mode == "finetune":
            pass
            # TODO


def off_diagonal(x):
    """Return a flattened view of the off-diagonal elements of a square matrix."""
    n, m = x.shape
    assert n == m
    return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()



from bdikit.value_matching.polyfuzz import (
    TFIDFValueMatcher,
    EditDistanceValueMatcher,
    FastTextValueMatcher,
    EmbeddingValueMatcher,
)


def test_textual_transformation_matching():
    threshold = 0.5
    for value_matcher in [
        TFIDFValueMatcher(threshold=threshold),
        EditDistanceValueMatcher(threshold=threshold),
    ]:
    
        # given
        source_values = ["Red Apple", "Banana", "Oorange", "dragon-fruits"]
        target_values = ["apple", "banana", "orange", "kiwi"]

        # when
        matches = value_matcher.match(source_values, target_values)

        # then
        assert len(matches) == 3

        mapped_matches = {match[0]: (match[1], match[2]) for match in matches}
        assert "dragon-fruits" not in mapped_matches
        assert mapped_matches["Red Apple"][0] == "apple"
        assert mapped_matches["Banana"][0] == "banana"
        assert mapped_matches["Oorange"][0] == "orange"

        scores = [match[2] for match in matches]
        assert all(score > threshold for score in scores)


def test_semantic_matching():
    threshold = 0.4
    value_matcher = FastTextValueMatcher(threshold=threshold)

    # given
    source_values = ["Computer", "Display", "Pencil"]
    target_values = ["PC", "Monitor", "Football field"]

    # when
    matches = value_matcher.match(source_values, target_values)

    # then
    assert len(matches) == 2

    mapped_matches = {match[0]: (match[1], match[2]) for match in matches}
    assert "Pencil" not in mapped_matches
    assert mapped_matches["Computer"][0] == "PC"
    assert mapped_matches["Display"][0] == "Monitor"

    scores = [match[2] for match in matches]
    assert all(score > threshold for score in scores)

    threshold = 0.6
    value_matcher = EmbeddingValueMatcher(threshold=threshold)

    # given
    source_values = ["Computer", "Display", "Pencil"]
    target_values = ["PC", "Monitor", "Football field"]

    # when
    matches = value_matcher.match(source_values, target_values)

    # then
    assert len(matches) == 2

    mapped_matches = {match[0]: (match[1], match[2]) for match in matches}
    assert "Pencil" not in mapped_matches
    assert mapped_matches["Computer"][0] == "PC"
    assert mapped_matches["Display"][0] == "Monitor"

    scores = [match[2] for match in matches]
    assert all(score > threshold for score in scores)


import pandas as pd
from typing import Dict, Callable
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from valentine import valentine_match
from valentine.algorithms.matcher_results import MatcherResults
from valentine.algorithms.jaccard_distance import StringDistanceFunction
from valentine.algorithms import (
    SimilarityFlooding,
    Coma,
    Cupid,
    DistributionBased,
    JaccardDistanceMatcher,
    BaseMatcher,
)


class ValentineSchemaMatcher(BaseSchemaMatcher):
    def __init__(self, matcher: BaseMatcher):
        self.matcher = matcher

    def map(self, source: pd.DataFrame, target: pd.DataFrame) -> Dict[str, str]:
        matches: MatcherResults = valentine_match(source, target, self.matcher)
        mappings = {}
        for match in matches.one_to_one():
            source_candidate = match[0][1]
            target_candidate = match[1][1]
            mappings[source_candidate] = target_candidate
        return self._fill_missing_matches(source, mappings)


class SimFloodSchemaMatcher(ValentineSchemaMatcher):
    def __init__(
        self, coeff_policy: str = "inverse_average", formula: str = "formula_c"
    ):
        super().__init__(SimilarityFlooding(coeff_policy=coeff_policy, formula=formula))


class ComaSchemaMatcher(ValentineSchemaMatcher):
    def __init__(
        self, max_n: int = 0, use_instances: bool = False, java_xmx: str = "1024m"
    ):
        super().__init__(
            Coma(max_n=max_n, use_instances=use_instances, java_xmx=java_xmx)
        )


class CupidSchemaMatcher(ValentineSchemaMatcher):
    def __init__(
        self,
        leaf_w_struct: float = 0.2,
        w_struct: float = 0.2,
        th_accept: float = 0.7,
        th_high: float = 0.6,
        th_low: float = 0.35,
        c_inc: float = 1.2,
        c_dec: float = 0.9,
        th_ns: float = 0.7,
        parallelism: int = 1,
    ):
        super().__init__(
            Cupid(
                leaf_w_struct=leaf_w_struct,
                w_struct=w_struct,
                th_accept=th_accept,
                th_high=th_high,
                th_low=th_low,
                c_inc=c_inc,
                c_dec=c_dec,
                th_ns=th_ns,
                parallelism=parallelism,
            )
        )


class DistributionBasedSchemaMatcher(ValentineSchemaMatcher):
    def __init__(
        self,
        threshold1: float = 0.15,
        threshold2: float = 0.15,
        quantiles: int = 256,
        process_num: int = 1,
    ):
        super().__init__(
            DistributionBased(
                threshold1=threshold1,
                threshold2=threshold2,
                quantiles=quantiles,
                process_num=process_num,
            )
        )


class JaccardSchemaMatcher(ValentineSchemaMatcher):
    def __init__(
        self,
        threshold_dist: float = 0.8,
        distance_fun: Callable[[str, str], float] = StringDistanceFunction.Levenshtein,
        process_num: int = 1,
    ):
        super().__init__(
            JaccardDistanceMatcher(
                threshold_dist=threshold_dist,
                distance_fun=distance_fun,
                process_num=process_num,
            )
        )



import pandas as pd
from typing import List, Dict


class BaseStandard:
    """
    Base class for all target standards, e.g. GDC.
    """

    def get_columns(self) -> List[str]:
        raise NotImplementedError("Subclasses must implement this method")

    def get_column_values(self, column_names: List[str]) -> Dict[str, List]:
        raise NotImplementedError("Subclasses must implement this method")

    def get_column_metadata(self, column_names: List[str]) -> Dict[str, Dict]:
        raise NotImplementedError("Subclasses must implement this method")

    def get_dataframe_rep(self) -> pd.DataFrame:
        raise NotImplementedError("Subclasses must implement this method")



import pandas as pd
import numpy as np
from typing import List
from bdikit.schema_matching.topk.base import (
    ColumnScore,
    TopkMatching,
    BaseTopkSchemaMatcher,
)
from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances
from bdikit.models.contrastive_learning.cl_api import (
    ContrastiveLearningAPI,
    DEFAULT_CL_MODEL,
)
from bdikit.models import ColumnEmbedder


class EmbeddingSimilarityTopkSchemaMatcher(BaseTopkSchemaMatcher):
    def __init__(self, column_embedder: ColumnEmbedder, metric: str = "cosine"):
        self.api = column_embedder
        self.metric = metric

    def get_recommendations(
        self, source: pd.DataFrame, target: pd.DataFrame, top_k: int = 10
    ) -> List[TopkMatching]:
        """
        Returns the top-k matching columns in the target table for each column
        in the source table. The ranking is based on the cosine similarity of
        the embeddings of the columns in the source and target tables.
        """
        l_features = self.api.get_embeddings(source)
        r_features = self.api.get_embeddings(target)
        if self.metric == "cosine":
            sim = cosine_similarity(l_features, r_features)  # type: ignore
        elif self.metric == "euclidean":
            sim = euclidean_distances(l_features, r_features)  # type: ignore
            sim = 1 / (1 + sim)
        else:
            raise ValueError(f"Invalid metric: {self.metric}")

        top_k_results = []
        for index, similarities in enumerate(sim):
            top_k_indices = np.argsort(similarities)[::-1][:top_k]
            top_k_columns = [
                ColumnScore(column_name=target.columns[i], score=similarities[i])
                for i in top_k_indices
            ]
            top_k_results.append(
                {
                    "source_column": source.columns[index],
                    "top_k_columns": top_k_columns,
                }
            )

        return top_k_results


class CLTopkSchemaMatcher(EmbeddingSimilarityTopkSchemaMatcher):
    def __init__(self, model_name: str = DEFAULT_CL_MODEL, metric: str = "cosine"):
        super().__init__(
            column_embedder=ContrastiveLearningAPI(model_name=model_name), metric=metric
        )



import importlib
from enum import Enum
from typing import Mapping, Any
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher


class SchemaMatchers(Enum):
    SIMFLOOD = (
        "similarity_flooding",
        "bdikit.schema_matching.one2one.valentine.SimFloodSchemaMatcher",
    )
    COMA = (
        "coma",
        "bdikit.schema_matching.one2one.valentine.ComaSchemaMatcher",
    )
    CUPID = (
        "cupid",
        "bdikit.schema_matching.one2one.valentine.CupidSchemaMatcher",
    )
    DISTRIBUTION_BASED = (
        "distribution_based",
        "bdikit.schema_matching.one2one.valentine.DistributionBasedSchemaMatcher",
    )
    JACCARD_DISTANCE = (
        "jaccard_distance",
        "bdikit.schema_matching.one2one.valentine.JaccardSchemaMatcher",
    )
    GPT = ("gpt", "bdikit.schema_matching.one2one.gpt.GPTSchemaMatcher")
    CT_LEARNING = (
        "ct_learning",
        "bdikit.schema_matching.one2one.contrastivelearning.ContrastiveLearningSchemaMatcher",
    )
    TWO_PHASE = (
        "two_phase",
        "bdikit.schema_matching.one2one.twophase.TwoPhaseSchemaMatcher",
    )
    MAX_VAL_SIM = (
        "max_val_sim",
        "bdikit.schema_matching.one2one.maxvalsim.MaxValSimSchemaMatcher",
    )

    MAGNETO_ZS_BP = (
        "magneto_zs_bp",
        "bdikit.schema_matching.topk.magneto.MagnetoZSBP",
    )

    MAGNETO_FT_BP = (
        "magneto_ft_bp",
        "bdikit.schema_matching.topk.magneto.MagnetoFTBP",
    )

    MAGNETO_ZS_LLM = (
        "magneto_zs_llm",
        "bdikit.schema_matching.topk.magneto.MagnetoZSLLM",
    )

    MAGNETO_FT_LLM = (
        "magneto_ft_llm",
        "bdikit.schema_matching.topk.magneto.MagnetoFTLLM",
    )

    def __init__(self, matcher_name: str, matcher_path: str):
        self.matcher_name = matcher_name
        self.matcher_path = matcher_path

    @staticmethod
    def get_matcher(
        matcher_name: str, **matcher_kwargs: Mapping[str, Any]
    ) -> BaseSchemaMatcher:
        if matcher_name not in matchers:
            names = ", ".join(list(matchers.keys()))
            raise ValueError(
                f"The {matcher_name} algorithm is not supported. "
                f"Supported algorithms are: {names}"
            )
        # Load the class dynamically
        module_path, class_name = matchers[matcher_name].rsplit(".", 1)
        module = importlib.import_module(module_path)

        return getattr(module, class_name)(**matcher_kwargs)


matchers = {method.matcher_name: method.matcher_path for method in SchemaMatchers}



import pandas as pd
from typing import Dict


class BaseSchemaMatcher:
    def map(self, source: pd.DataFrame, target: pd.DataFrame) -> Dict[str, str]:
        raise NotImplementedError("Subclasses must implement this method")

    def _fill_missing_matches(
        self, dataset: pd.DataFrame, matches: Dict[str, str]
    ) -> Dict[str, str]:
        for column in dataset.columns:
            if column not in matches:
                matches[column] = ""
        return matches



import pandas as pd
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from bdikit.models.contrastive_learning.cl_api import DEFAULT_CL_MODEL
from bdikit.schema_matching.topk.contrastivelearning import CLTopkSchemaMatcher


class ContrastiveLearningSchemaMatcher(BaseSchemaMatcher):
    def __init__(self, model_name: str = DEFAULT_CL_MODEL):
        self.topk_matcher = CLTopkSchemaMatcher(model_name=model_name)

    def map(self, source: pd.DataFrame, target: pd.DataFrame):
        topk_matches = self.topk_matcher.get_recommendations(source, target, top_k=1)
        matches = {}
        for column, top_k_match in zip(source.columns, topk_matches):
            candidate = top_k_match["top_k_columns"][0][0]
            if candidate in target.columns:
                matches[column] = candidate
        return self._fill_missing_matches(source, matches)



from typing import List, Dict, Tuple, Optional
from bdikit.config import get_device
import numpy as np
import pandas as pd
import torch
from bdikit.models.contrastive_learning.cl_models import (
    BarlowTwinsSimCLR,
)
from bdikit.models.contrastive_learning.cl_pretrained_dataset import (
    PretrainTableDataset,
)
from sklearn.metrics.pairwise import cosine_similarity
from tqdm.auto import tqdm
from bdikit.download import get_cached_model_or_download
from bdikit.utils import check_embedding_cache, write_embeddings_to_cache
from bdikit.models import ColumnEmbedder


DEFAULT_CL_MODEL = "bdi-cl-v0.2"


class ContrastiveLearningAPI(ColumnEmbedder):
    def __init__(
        self,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        batch_size: int = 128,
    ):
        if model_name and model_path:
            raise ValueError(
                "Only one of model_name or model_path should be provided "
                "(they are mutually exclusive)"
            )

        if model_path:
            self.model_path = model_path
        elif model_name:
            self.model_path = get_cached_model_or_download(model_name)
        else:
            raise ValueError("Either model_name or model_path must be provided")

        self.unlabeled = PretrainTableDataset()
        self.batch_size = batch_size
        self.device = get_device()
        self.model = self.load_checkpoint()

    def load_checkpoint(self, lm: str = "roberta"):
        ckpt = torch.load(self.model_path, map_location=torch.device("cpu"))
        scale_loss = 0.1
        lambd = 3.9
        model = BarlowTwinsSimCLR(scale_loss, lambd, device=self.device, lm=lm)
        model = model.to(self.device)
        model.load_state_dict(ckpt["model"])

        return model

    def get_embeddings(self, table: pd.DataFrame) -> List[np.ndarray]:
        return self._load_table_tokens(table)

    def get_recommendations(
        self, table: pd.DataFrame, target: pd.DataFrame, top_k: int = 10
    ) -> Tuple[List, List[Dict]]:

        l_features = self._load_table_tokens(table)
        r_features = self._load_table_tokens(target)
        cosine_sim = cosine_similarity(l_features, r_features)

        # print(f"l_features - {len(l_features)}:{l_features[0].shape}\nr-feature - {len(r_features)}:{r_features[0].shape}\nCosine - {cosine_sim.shape}")

        top_k_results = []
        l_column_ids = table.columns
        gt_column_ids = target.columns

        for index, similarities in enumerate(cosine_sim):
            top_k_indices = np.argsort(similarities)[::-1][:top_k]
            top_k_column_names = [gt_column_ids[i] for i in top_k_indices]
            top_k_similarities = [str(round(similarities[i], 4)) for i in top_k_indices]
            top_k_columns = list(zip(top_k_column_names, top_k_similarities))
            result = {
                "Candidate column": l_column_ids[index],
                "Top k columns": top_k_columns,
            }
            top_k_results.append(result)
        recommendations = self._extract_recommendations_from_top_k(top_k_results)
        return recommendations, top_k_results

    def _extract_recommendations_from_top_k(self, top_k_results: List[dict]):
        recommendations = set()
        for result in top_k_results:
            for name, _ in result["Top k columns"]:
                recommendations.add(name)
        return list(recommendations)

    def _sample_to_15_rows(self, table: pd.DataFrame):
        if len(table) > 15:
            unique_rows = table.drop_duplicates()
            num_unique_rows = len(unique_rows)
            if num_unique_rows <= 15:
                needed_rows = 15 - num_unique_rows
                additional_rows = table[~table.index.isin(unique_rows.index)].sample(
                    n=needed_rows, replace=True, random_state=1
                )
                table = pd.concat([unique_rows, additional_rows])
            else:
                table = unique_rows.sample(n=15, random_state=1)
        return table

    def _load_table_tokens(self, table: pd.DataFrame) -> List[np.ndarray]:

        embedding_file, embeddings = check_embedding_cache(table, self.model_path)

        if embeddings != None:
            print(f"Table features loaded for {len(table.columns)} columns")
            return embeddings

        print(f"Extracting features from {len(table.columns)} columns...")
        tables = []
        for _, column in enumerate(table.columns):
            curr_table = pd.DataFrame(table[column])
            curr_table = self._sample_to_15_rows(curr_table)
            tables.append(curr_table)
        vectors = self._inference_on_tables(tables)
        embeddings = [vec[-1] for vec in vectors]

        if embedding_file != None:
            write_embeddings_to_cache(embedding_file, embeddings)

        return embeddings

    def _inference_on_tables(self, tables: List[pd.DataFrame]) -> List[List]:
        total = len(tables)
        batch = []
        results = []

        for tid, table in tqdm(enumerate(tables), total=total):
            x, _ = self.unlabeled._tokenize(table)
            batch.append((x, x, []))

            if tid == total - 1 or len(batch) == self.batch_size:
                with torch.no_grad():
                    x, _, _ = self.unlabeled.pad(batch)
                    column_vectors = self.model.inference(x)
                    ptr = 0
                    for xi in x:
                        current = []
                        for token_id in xi:
                            if token_id == self.unlabeled.tokenizer.cls_token_id:
                                current.append(column_vectors[ptr].cpu().numpy())
                                ptr += 1
                        results.append(current)
                batch.clear()
        return results



import pandas as pd
from typing import Optional
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from bdikit.models.contrastive_learning.cl_api import DEFAULT_CL_MODEL
from bdikit.schema_matching.topk.base import BaseTopkSchemaMatcher
from bdikit.schema_matching.topk.contrastivelearning import CLTopkSchemaMatcher
from bdikit.value_matching.polyfuzz import TFIDFValueMatcher
from bdikit.value_matching.base import BaseValueMatcher


class MaxValSimSchemaMatcher(BaseSchemaMatcher):
    def __init__(
        self,
        top_k: int = 20,
        top_k_matcher: Optional[BaseTopkSchemaMatcher] = None,
        value_matcher: Optional[BaseValueMatcher] = None,
    ):
        if top_k_matcher is None:
            self.api = CLTopkSchemaMatcher(DEFAULT_CL_MODEL)
        elif isinstance(top_k_matcher, BaseTopkSchemaMatcher):
            self.api = top_k_matcher
        else:
            raise ValueError(
                f"Invalid top_k_matcher type: {type(top_k_matcher)}. "
                "Must be a subclass of {BaseTopkColumnMatcher.__name__}"
            )

        if value_matcher is None:
            self.value_matcher = TFIDFValueMatcher()
        elif isinstance(value_matcher, BaseValueMatcher):
            self.value_matcher = value_matcher
        else:
            raise ValueError(
                f"Invalid value_matcher type: {type(value_matcher)}. "
                "Must be a subclass of {BaseValueMatcher.__name__}"
            )

        self.top_k = top_k

    def unique_string_values(self, column: pd.Series) -> pd.Series:
        column = column.dropna()
        if pd.api.types.is_string_dtype(column):
            return pd.Series(column.unique(), name=column.name)
        else:
            return pd.Series(column.unique().astype(str), name=column.name)

    def map(
        self,
        source: pd.DataFrame,
        target: pd.DataFrame,
    ):
        topk_column_matches = self.api.get_recommendations(source, target, self.top_k)

        matches = {}
        for source_column_name, scope in zip(source.columns, topk_column_matches):

            source_column_name = scope["source_column"]
            top_k_columns = scope["top_k_columns"]

            source_column = source[source_column_name]

            if not pd.api.types.is_string_dtype(source_column):
                matches[source_column_name] = top_k_columns[0].column_name
                continue

            source_values = self.unique_string_values(source_column).to_list()

            scores = []
            for top_column in top_k_columns:
                target_column_name = top_column.column_name
                target_column = target[target_column_name]
                target_values = self.unique_string_values(target_column).to_list()
                value_matches = self.value_matcher.match(source_values, target_values)
                score = sum([m.similarity for m in value_matches]) / len(target_values)
                score = (top_column.score + score) / 2.0
                scores.append((source_column_name, target_column_name, score))

            sorted_columns = sorted(scores, key=lambda it: it[2], reverse=True)

            matches[source_column_name] = sorted_columns[0][1]

        return self._fill_missing_matches(source, matches)



import os
import hashlib
import pandas as pd
from os.path import join, dirname, isfile
from bdikit.download import BDIKIT_EMBEDDINGS_CACHE_DIR


def hash_dataframe(df: pd.DataFrame) -> str:
    hash_object = hashlib.sha256()

    columns_string = ",".join(df.columns) + "\n"
    hash_object.update(columns_string.encode())

    for row in df.itertuples(index=False, name=None):
        row_string = ",".join(map(str, row)) + "\n"
        hash_object.update(row_string.encode())

    return hash_object.hexdigest()


def write_embeddings_to_cache(embedding_file: str, embeddings: list):

    os.makedirs(dirname(embedding_file), exist_ok=True)

    with open(embedding_file, "w") as file:
        for vec in embeddings:
            file.write(",".join([str(val) for val in vec]) + "\n")


def check_embedding_cache(table: pd.DataFrame, model_path: str):
    embedding_file = None
    embeddings = None
    table_hash = hash_dataframe(table)
    model_name = model_path.split("/")[-1]
    cache_model_path = join(BDIKIT_EMBEDDINGS_CACHE_DIR, model_name)
    os.makedirs(cache_model_path, exist_ok=True)

    hash_list = {
        f for f in os.listdir(cache_model_path) if isfile(join(cache_model_path, f))
    }

    embedding_file = join(cache_model_path, table_hash)

    # Check if table for computing embedding is the same as the tables we have in resources
    if table_hash in hash_list:
        if isfile(embedding_file):
            try:
                # Load embeddings from disk
                with open(embedding_file, "r") as file:
                    embeddings = [
                        [float(val) for val in vec.split(",")]
                        for vec in file.read().split("\n")
                        if vec.strip()
                    ]

            except Exception as e:
                print(f"Error loading features from cache: {e}")
                embeddings = None

    return embedding_file, embeddings



import bdikit as bdi
import numpy as np
import pandas as pd
import numpy as np
from bdikit.mapping_functions import (
    FunctionValueMapper,
    IdentityValueMapper,
)


def test_bdi_match_schema_with_dataframes():
    # given
    source = pd.DataFrame({"column_1": ["a1", "b1", "c1"], "col_2": ["a2", "b2", "c2"]})
    target = pd.DataFrame({"column_1a": ["a1", "b1", "c1"], "col2": ["a2", "b2", "c2"]})

    # when
    df_matches = bdi.match_schema(source, target=target, method="similarity_flooding")

    # then assert that the df_matches contains a row with the value 'column_1'
    # in the column 'source' and the value 'column_1a' in the 'target' value
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns

    df_filter = df_matches["source"] == "column_1"
    assert df_matches[df_filter]["target"].values[0] == "column_1a"

    df_filter = df_matches["source"] == "col_2"
    assert df_matches[df_filter]["target"].values[0] == "col2"


def test_bdi_match_schema_to_gdc():
    # given
    source = pd.DataFrame(
        {
            "FIGO_stage": [
                "Stage 0",
                "Stage I",
                "Stage IA",
                "Stage IA1",
                "Stage IA2",
            ],
            "Ethnicity": [
                "Not-Hispanic or Latino",
                "Hispanic or Latino",
                "Not reported",
                "Hispanic or Latino",
                "Hispanic or Latino",
            ],
        }
    )

    # when
    df_matches = bdi.match_schema(source, target="gdc", method="coma")

    # then df_matches must contain target columns that come from the GDC dictionary
    assert df_matches.empty == False
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns

    df_filter = df_matches["source"] == "Ethnicity"
    assert df_matches[df_filter]["target"].values[0] == "ethnicity"

    df_filter = df_matches["source"] == "FIGO_stage"
    assert df_matches[df_filter]["target"].values[0] == "figo_stage"


def test_bdi_top_matches_with_dataframes():
    # given
    source = pd.DataFrame({"tumor_size": ["a1", "b1", "c1"]})
    target = pd.DataFrame(
        {
            "color": ["red", "blue", "green", "yellow"],
            "tumor_length": [12, 23, 34, 45],
            "tumor_width": [12, 23, 34, 45],
            "tumor_size": [12, 23, 34, 45],
        }
    )

    #
    # First test with method ct_learning and default args
    #

    # when
    df_matches = bdi.top_matches(
        source,
        target=target,
        top_k=3,
        method="ct_learning",
    )

    # then
    assert len(df_matches.index) == 3
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns
    assert "similarity" in df_matches.columns

    df_filter = df_matches["source"] == "tumor_size"
    assert "tumor_size" in df_matches[df_filter]["target"].tolist()
    assert "tumor_width" in df_matches[df_filter]["target"].tolist()
    assert "tumor_length" in df_matches[df_filter]["target"].tolist()

    #
    # Now test with ct_learning and euclidean distance
    #

    # when
    df_matches = bdi.top_matches(
        source,
        target=target,
        top_k=3,
        method="ct_learning",
        method_args={"metric": "euclidean"},
    )

    # then
    assert len(df_matches.index) == 3
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns
    assert "similarity" in df_matches.columns

    df_filter = df_matches["source"] == "tumor_size"
    assert "tumor_size" in df_matches[df_filter]["target"].tolist()
    assert "tumor_width" in df_matches[df_filter]["target"].tolist()
    assert "tumor_length" in df_matches[df_filter]["target"].tolist()


def test_bdi_top_matches_gdc():
    # given
    source = pd.DataFrame(
        {
            "FIGO_stage": [
                "Stage 0",
                "Stage I",
                "Stage IA",
                "Stage IA1",
                "Stage IA2",
            ],
            "Ethnicity": [
                "Not-Hispanic or Latino",
                "Hispanic or Latino",
                "Not reported",
                "Hispanic or Latino",
                "Hispanic or Latino",
            ],
        }
    )

    # when
    df_matches = bdi.top_matches(source, target="gdc", top_k=5)

    # then
    assert df_matches.empty == False
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns
    assert "similarity" in df_matches.columns

    df_filter = df_matches["source"] == "FIGO_stage"
    assert len(df_matches[df_filter]) == 5
    assert "figo_stage" in df_matches[df_filter]["target"].tolist()
    assert "uicc_clinical_stage" in df_matches[df_filter]["target"].tolist()

    df_filter = df_matches["source"] == "Ethnicity"
    assert len(df_matches[df_filter]) == 5
    assert "ethnicity" in df_matches[df_filter]["target"].tolist()
    assert "race" in df_matches[df_filter]["target"].tolist()


def test_map_dataframe_column_values():
    # given
    str_column_1 = ["a", "b", "c", "d", "e"]
    str_column_2 = ["a", "b", "c", "d", "e"]
    df_base = pd.DataFrame({"column_str_1": str_column_1, "column_str_2": str_column_2})

    value_mapping_spec = [
        {
            "source": "column_str_1",
            "target": "string column 1",
            "mapper": IdentityValueMapper(),
        },
        {
            "source": "column_str_2",
            "target": "string column 2",
            "mapper": FunctionValueMapper(function=lambda x: x.upper()),
        },
    ]

    # when
    df_mapped = bdi.materialize_mapping(df_base, mapping_spec=value_mapping_spec)

    # then
    assert len(df_mapped.columns) == 2

    assert "string column 1" in df_mapped.columns
    assert df_mapped["string column 1"].eq(str_column_1).all()

    assert "string column 2" in df_mapped.columns
    assert df_mapped["string column 2"].eq(["A", "B", "C", "D", "E"]).all()


def test_value_mapping_dataframe():
    # given
    df_source = pd.DataFrame(
        {"src_column": ["Red Apple", "Banana", "Oorange", "Strawberry"]}
    )
    df_target = pd.DataFrame(
        {"tgt_column": ["apple", "banana", "orange", "kiwi", "grapes"]}
    )

    df_matches = pd.DataFrame({"source": ["src_column"], "target": ["tgt_column"]})

    # when
    value_mappings = bdi.match_values(df_source, df_target, df_matches, method="tfidf")

    # then
    assert len(value_mappings) == 1
    mapping = value_mappings[0]
    assert mapping is not None
    assert isinstance(mapping, pd.DataFrame)
    assert mapping.attrs["source"] == "src_column"
    assert mapping.attrs["target"] == "tgt_column"
    assert len(mapping) == len(df_source)


def test_end_to_end_api_integration():
    # given
    df_source = pd.DataFrame(
        {"src_column": ["Red Apple", "Banana", "Oorange", "Strawberry"]}
    )
    df_target = pd.DataFrame(
        {"tgt_column": ["apple", "banana", "orange", "kiwi", "grapes"]}
    )

    # when
    column_mappings = bdi.match_schema(df_source, df_target, method="coma")
    # then
    assert column_mappings is not None
    assert column_mappings.empty == False
    assert "source" in column_mappings.columns
    assert "target" in column_mappings.columns
    assert len(column_mappings.index) == 1

    # when: pass output of match_schema() directly to materialize_mapping(),
    # the column must be renamed to the target column without any value mapping
    df_mapped = bdi.materialize_mapping(df_source, column_mappings)
    # then
    assert "tgt_column" in df_mapped.columns
    assert df_mapped["tgt_column"].tolist() == [
        "Red Apple",
        "Banana",
        "Oorange",
        "Strawberry",
    ]

    # when: we pass the output of match_schema()
    value_mappings = bdi.match_values(
        df_source, df_target, column_mappings, method="tfidf"
    )

    # then: a list of value matches must be computed
    assert len(value_mappings) == 1
    mapping = value_mappings[0]
    assert mapping is not None
    assert isinstance(mapping, pd.DataFrame)
    assert len(mapping) == len(df_source)
    assert mapping.attrs["source"] == "src_column"
    assert mapping.attrs["target"] == "tgt_column"

    # when: pass output of match_values() to materialize_mapping(),
    df_mapped = bdi.materialize_mapping(df_source, value_mappings)

    # then: the column must be ranamed and values must be mapped to the
    # matching values found during the value matching step
    assert isinstance(df_mapped, pd.DataFrame)
    assert "tgt_column" in df_mapped.columns
    assert df_mapped["tgt_column"].tolist() == ["apple", "banana", "orange", np.nan]

    # when: pass output of match_values() to merge_mappings() and then to
    # materialize_mapping()

    harmonization_spec = bdi.merge_mappings(value_mappings, [])
    df_mapped = bdi.materialize_mapping(df_source, harmonization_spec)

    # then: the column must be renamed and values must be mapped
    assert isinstance(df_mapped, pd.DataFrame)
    assert "tgt_column" in df_mapped.columns
    assert df_mapped["tgt_column"].tolist() == ["apple", "banana", "orange", np.nan]

    # when: user mappings are specified in merge_mappings()
    user_mappings = [
        {
            "source": "src_column",
            "target": "tgt_column",
            "matches": [
                ("Red Apple", "APPLE"),
                ("Banana", "BANANA"),
                ("Oorange", "ORANGE"),
            ],
        }
    ]
    harmonization_spec = bdi.merge_mappings(value_mappings, user_mappings)
    df_mapped = bdi.materialize_mapping(df_source, harmonization_spec)

    # then: user mappings take precedence, so the column must be renamed and
    # values must be mapped according the provide user_mappings
    assert "tgt_column" in df_mapped.columns
    assert df_mapped["tgt_column"].tolist() == ["APPLE", "BANANA", "ORANGE", np.nan]


def test_top_matches_and_match_values_integration():
    # given
    df_source = pd.DataFrame(
        {"fruits": ["Red Apple", "Banana", "Oorange", "Strawberry"]}
    )
    df_target = pd.DataFrame(
        {
            "fruit_types": ["apple", "banana", "orange", "kiwi", "grapes"],
            "fruit_names": ["apple", "banana", "melon", "kiwi", "grapes"],
            "fruit_id": ["1", "2", "3", "4", "5"],
        }
    )

    # when
    df_matches = bdi.top_matches(df_source, target=df_target)

    # then
    assert len(df_matches.index) == 3
    assert "source" in df_matches.columns
    assert "target" in df_matches.columns
    assert "similarity" in df_matches.columns

    # when
    df_matches = bdi.match_values(
        df_source, df_target, column_mapping=df_matches, method="tfidf"
    )
    assert isinstance(df_matches, list)
    assert len(df_matches) == 3
    for df in df_matches:
        assert isinstance(df, pd.DataFrame)
        assert "source" in df.columns
        assert "target" in df.columns
        assert "similarity" in df.columns
        assert df.attrs["source"] == "fruits"
        assert df.attrs["target"] in ["fruit_types", "fruit_names", "fruit_id"]


def test_top_value_matches():
    # given
    df_source = pd.DataFrame({"fruits": ["Applee", "Bananaa", "Oorange", "Strawberry"]})
    df_target = pd.DataFrame(
        {
            "fruit_names": [
                "apple",
                "red apple",
                "banana",
                "mx banana",
                "melon",
                "kiwi",
                "grapes",
            ],
            "fruit_id": ["1", "2", "3", "4", "5", "6", "7"],
        }
    )
    column_mapping = ("fruits", "fruit_names")
    # when
    matches = bdi.top_value_matches(df_source, df_target, column_mapping)

    # then
    assert len(matches) == 4  # number of dataframes in the list

    # when
    df_match = matches[0]  # top matches for apple

    # then
    assert len(df_match) == 2
    assert "source" in df_match.columns
    assert "target" in df_match.columns
    assert "similarity" in df_match.columns

    # when
    df_match = matches[1]  # top matches for banana

    # then
    assert len(df_match) == 2
    assert "source" in df_match.columns
    assert "target" in df_match.columns
    assert "similarity" in df_match.columns

    # when
    df_match = matches[2]  # top matches for orange

    # then
    assert len(df_match) == 1
    assert "source" in df_match.columns
    assert "target" in df_match.columns
    assert "similarity" in df_match.columns

def test_preview_domain():
    # given
    source = pd.DataFrame(
        {
            "name": ["John Doe", "Jane Doe", "Alice Smith", "Bob Smith"],
            "age": [30, 25, 45, 35],
        }
    )

    # when
    preview = bdi.preview_domain(source, "age")

    # then
    # preview must contain only the column "value_name" and the unique 
    # values of the column "age"
    assert preview is not None
    assert isinstance(preview, pd.DataFrame)
    assert "value_name" in preview.columns
    assert "column_description" not in preview.columns
    assert "value_description" not in preview.columns
    assert source["age"].eq(preview["value_name"]).all()

    # when
    preview = bdi.preview_domain("gdc", "age_at_diagnosis")

    # then
    # preview must contain only the column "column_description" since there
    # are sample values in the GDC dictionary
    assert preview is not None
    assert isinstance(preview, pd.DataFrame)
    assert "value_name" not in preview.columns
    assert "value_description" not in preview.columns
    assert "column_description" in preview.columns



import pandas as pd
from typing import Optional
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from bdikit.schema_matching.one2one.valentine import SimFloodSchemaMatcher
from bdikit.models.contrastive_learning.cl_api import DEFAULT_CL_MODEL
from bdikit.schema_matching.topk.base import BaseTopkSchemaMatcher
from bdikit.schema_matching.topk.contrastivelearning import CLTopkSchemaMatcher


class TwoPhaseSchemaMatcher(BaseSchemaMatcher):
    def __init__(
        self,
        top_k: int = 20,
        top_k_matcher: Optional[BaseTopkSchemaMatcher] = None,
        schema_matcher: BaseSchemaMatcher = SimFloodSchemaMatcher(),
    ):
        if top_k_matcher is None:
            self.api = CLTopkSchemaMatcher(DEFAULT_CL_MODEL)
        elif isinstance(top_k_matcher, BaseTopkSchemaMatcher):
            self.api = top_k_matcher
        else:
            raise ValueError(
                f"Invalid top_k_matcher type: {type(top_k_matcher)}. "
                "Must be a subclass of {BaseTopkColumnMatcher.__name__}"
            )

        self.schema_matcher = schema_matcher
        self.top_k = top_k

    def map(
        self,
        source: pd.DataFrame,
        target: pd.DataFrame,
    ):
        topk_column_matches = self.api.get_recommendations(source, target, self.top_k)

        matches = {}
        for column, scope in zip(source.columns, topk_column_matches):
            candidates = [
                cand[0] for cand in scope["top_k_columns"] if cand[0] in target.columns
            ]
            reduced_source = source[[column]]
            reduced_target = target[candidates]
            partial_matches = self.schema_matcher.map(reduced_source, reduced_target)

            if column in partial_matches:
                matches[column] = partial_matches[column]

        return self._fill_missing_matches(source, matches)



import pandas as pd
import numpy as np
from typing import Any, Callable
from collections import defaultdict


class ValueMapper:
    """
    A ValueMapper represents objects that transform the values in a input
    column to the values from a new output column.
    """

    def map(self, input_column: pd.Series) -> pd.Series:
        """
        Every concrete ValueMapper should implement this method, which takes a
        pandas Series as input and returns a new pandas Series with transformed
        values.
        """
        pass


class IdentityValueMapper(ValueMapper):
    """
    A column mapper that maps each value in input column into itself.
    """

    def map(self, input_column: pd.Series) -> pd.Series:
        """
        Simply copies the values in input_column to the output column.
        """
        return input_column.copy()


class FunctionValueMapper(ValueMapper):
    """
    A column mapper that transforms each value in the input column using the
    provided custom function.
    """

    def __init__(self, function: Callable[[pd.Series], pd.Series]):
        self.function = function

    def map(self, input_column: pd.Series) -> pd.Series:
        """
        Applies the given function to each value in input_column to generate
        the output column.
        """
        return input_column.map(self.function)


class DictionaryMapper(ValueMapper):
    """
    A column mapper that transforms each value in the input column using the
    values stored in the provided dictionary.
    """

    def __init__(self, dictionary: dict, missing_key_value: Any = np.nan):
        self.dictionary = defaultdict(lambda: missing_key_value, dictionary)

    def map(self, input_column: pd.Series) -> pd.Series:
        """
        Transforms the values in the input_column to the values specified in
        the dictionary provided using the object constructor.
        """
        return input_column.map(self.dictionary, na_action=None)



import pandas as pd
from typing import Dict, Any, List
from magneto import Magneto as Magneto_Lib
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher
from bdikit.download import get_cached_model_or_download
from bdikit.schema_matching.topk.base import (
    ColumnScore,
    TopkMatching,
    BaseTopkSchemaMatcher,
)

DEFAULT_MAGNETO_MODEL = "magneto-gdc-v0.1"


class MagnetoBase(BaseSchemaMatcher, BaseTopkSchemaMatcher):
    def __init__(self, kwargs: Dict[str, Any] = None):
        if kwargs is None:
            kwargs = {}
        self.magneto = Magneto_Lib(**kwargs)

    def map(
        self,
        source: pd.DataFrame,
        target: pd.DataFrame,
    ):
        # There is an issue in Magneto to get the top-1 match, so get top 2 and then filter
        self.magneto.params["topk"] = 2  # Magneto does not provide a method to set topk
        raw_matches = self.magneto.get_matches(source, target)

        # Organizing data into the desired structure
        sorted_dict = {}
        for (source, target), score in raw_matches.items():
            source_column = source[1]
            target_column = target[1]
            if source_column not in sorted_dict:
                sorted_dict[source_column] = []
            sorted_dict[source_column].append((target_column, score))

        # Sorting the lists by value in descending order and get top 1
        formatted_matches = {}
        for key in sorted_dict:
            sorted_matches = sorted(sorted_dict[key], key=lambda x: x[1], reverse=True)
            formatted_matches[key] = sorted_matches[0][0]

        return formatted_matches

    def get_recommendations(
        self, source: pd.DataFrame, target: pd.DataFrame, top_k: int
    ) -> List[TopkMatching]:
        self.magneto.params["topk"] = (
            top_k  # Magneto does not provide a method to set topk
        )
        raw_matches = self.magneto.get_matches(source, target)

        # Organizing data into the desired structure
        sorted_dict = {}
        for (source, target), score in raw_matches.items():
            source_column = source[1]
            target_column = target[1]
            if source_column not in sorted_dict:
                sorted_dict[source_column] = []
            sorted_dict[source_column].append((target_column, score))

        # Sorting the lists by value in descending order and format top k
        top_k_results = []
        for key in sorted_dict:
            sorted_matches = sorted(sorted_dict[key], key=lambda x: x[1], reverse=True)
            top_k_columns = [ColumnScore(name, score) for name, score in sorted_matches]
            top_k_results.append(
                {
                    "source_column": [key] * len(top_k_columns),
                    "top_k_columns": top_k_columns,
                }
            )

        return top_k_results


class MagnetoZSBP(MagnetoBase):
    def __init__(self):
        super().__init__()


class MagnetoFTBP(MagnetoBase):
    def __init__(
        self,
        encoding_mode: str = "header_values_verbose",
        model_name: str = DEFAULT_MAGNETO_MODEL,
        model_path: str = None,
    ):
        embedding_model = check_magneto_model(model_name, model_path)
        kwargs = {"encoding_mode": encoding_mode, "embedding_model": embedding_model}
        super().__init__(kwargs)


class MagnetoZSLLM(MagnetoBase):
    def __init__(self):
        kwargs = {"use_bp_reranker": False, "use_gpt_reranker": True}
        super().__init__(kwargs)


class MagnetoFTLLM(MagnetoBase):
    def __init__(
        self,
        encoding_mode: str = "header_values_verbose",
        model_name: str = DEFAULT_MAGNETO_MODEL,
        model_path: str = None,
    ):
        embedding_model = check_magneto_model(model_name, model_path)
        kwargs = {
            "encoding_mode": encoding_mode,
            "embedding_model": embedding_model,
            "use_bp_reranker": False,
            "use_gpt_reranker": True,
        }
        super().__init__(kwargs)


def check_magneto_model(model_name: str, model_path: str):
    if model_name and model_path:
        raise ValueError(
            "Only one of model_name or model_path should be provided "
            "(they are mutually exclusive)"
        )

    if model_path:
        return model_path
    elif model_name:
        return get_cached_model_or_download(model_name)
    else:
        raise ValueError("Either model_name or model_path must be provided")



import os
import sys
import requests
from tqdm.auto import tqdm


default_os_cache_dir = os.getenv(
    "XDG_CACHE_HOME", os.path.join(os.path.expanduser("~"), ".cache")
)
BDIKIT_CACHE_DIR = os.getenv(
    "BDIKIT_CACHE", os.path.join(default_os_cache_dir, "bdikit")
)
BDIKIT_MODELS_CACHE_DIR = os.path.join(BDIKIT_CACHE_DIR, "models")

BUILTIN_MODELS_BOX_URL = {
    "cl-reducer-v0.1": "https://nyu.box.com/shared/static/hc4qxzbuxz0uoynfwy4pe2yxo5ch6xgm.pt",
    "bdi-cl-v0.2": "https://nyu.box.com/shared/static/1vdc28kzbpoj6ey95bksaww541p9gj31.pt",
    "magneto-gdc-v0.1": "https://nyu.box.com/shared/static/140g2rq1izc1wqs1ssrml6jzag3qa0mu.pth",
}

BDIKIT_EMBEDDINGS_CACHE_DIR = os.path.join(BDIKIT_CACHE_DIR, "embeddings")


def download_file_url(url: str, destination: str):
    # start the download stream
    response = requests.get(url, stream=True)
    # read total sizes in bytes from http headers
    total_size = int(response.headers.get("content-length", 0))
    # download the file in chunks and write to destination file
    block_size = 1024
    with tqdm(total=total_size, unit="B", unit_scale=True) as progress_bar:
        with open(destination, "wb") as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
    # check if filed was downloaded completely
    if total_size != 0 and progress_bar.n != total_size:
        raise RuntimeError(
            f"Failed to download file (expected {total_size} bytes, got {progress_bar.n} bytes)"
        )


def get_cache_file_path(model_name: str):
    if not os.path.exists(BDIKIT_MODELS_CACHE_DIR):
        print(
            f"Cache directory does not exist, creating it at: {BDIKIT_MODELS_CACHE_DIR}"
        )
        os.makedirs(BDIKIT_MODELS_CACHE_DIR)
    return os.path.join(BDIKIT_MODELS_CACHE_DIR, model_name)


def get_cached_model_or_download(model_name: str):
    model_path = get_cache_file_path(model_name)
    if not os.path.exists(model_path):
        if model_name in BUILTIN_MODELS_BOX_URL:
            print(f"Downloading model {model_name} from Box to {model_path}")
            download_file_url(BUILTIN_MODELS_BOX_URL[model_name], model_path)
        else:
            raise ValueError(f"Model {model_name} not found in builtin models")
    return model_path


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Please provide a model_id as a command line argument.")
        sys.exit(1)

    model_id = sys.argv[1]
    model_path = get_cached_model_or_download(model_id)
    print(f"Downloaded model: {model_path}")



import importlib
from enum import Enum
from typing import Mapping, Any
from bdikit.value_matching.base import BaseValueMatcher


class ValueMatchers(Enum):
    TFIDF = ("tfidf", "bdikit.value_matching.polyfuzz.TFIDFValueMatcher")
    EDIT = (
        "edit_distance",
        "bdikit.value_matching.polyfuzz.EditDistanceValueMatcher",
    )
    EMBEDDINGS = (
        "embedding",
        "bdikit.value_matching.polyfuzz.EmbeddingValueMatcher",
    )
    FASTTEXT = (
        "fasttext",
        "bdikit.value_matching.polyfuzz.FastTextValueMatcher",
    )
    GPT = ("gpt", "bdikit.value_matching.gpt.GPTValueMatcher")

    def __init__(self, matcher_name: str, matcher_path: str):
        self.matcher_name = matcher_name
        self.matcher_path = matcher_path

    @staticmethod
    def get_matcher(
        matcher_name: str, **matcher_kwargs: Mapping[str, Any]
    ) -> BaseValueMatcher:
        if matcher_name not in matchers:
            names = ", ".join(list(matchers.keys()))
            raise ValueError(
                f"The {matcher_name} algorithm is not supported. "
                f"Supported algorithms are: {names}"
            )
        # Load the class dynamically
        module_path, class_name = matchers[matcher_name].rsplit(".", 1)
        module = importlib.import_module(module_path)

        return getattr(module, class_name)(**matcher_kwargs)


matchers = {method.matcher_name: method.matcher_path for method in ValueMatchers}



from typing import List
from openai import OpenAI
from bdikit.value_matching.base import BaseValueMatcher, ValueMatch
from bdikit.config import VALUE_MATCHING_THRESHOLD


class GPTValueMatcher(BaseValueMatcher):
    def __init__(
        self,
        threshold: float = VALUE_MATCHING_THRESHOLD,
    ):
        self.client = OpenAI()
        self.threshold = threshold

    def match(
        self,
        source_values: List[str],
        target_values: List[str],
    ) -> List[ValueMatch]:
        target_values_set = set(target_values)
        matches = []

        for source_value in source_values:
            completion = self.client.chat.completions.create(
                model="gpt-4-turbo-preview",
                messages=[
                    {
                        "role": "system",
                        "content": "You are an intelligent system that given a term, you have to choose a value from a list that best matches the term. "
                        "These terms belong to the medical domain, and the list contains terms in the Genomics Data Commons (GDC) format.",
                    },
                    {
                        "role": "user",
                        "content": f'For the term: "{source_value}", choose a value from this list {target_values}. '
                        "Return the value from the list with a similarity score, between 0 and 1, with 1 indicating the highest similarity. "
                        "DO NOT PROVIDE ANY OTHER OUTPUT TEXT OR EXPLANATION. "
                        'Only provide a Python dictionary. For example {"term": "term from the list", "score": 0.8}.',
                    },
                ],
            )

            response_message = completion.choices[0].message.content
            try:
                response_dict = ast.literal_eval(response_message)
                target_value = response_dict["term"]
                score = float(response_dict["score"])
                if target_value in target_values_set and score >= self.threshold:
                    matches.append(ValueMatch(source_value, target_value, score))
            except:
                print(
                    f'Errors parsing response for "{source_value}": {response_message}'
                )

        return matches



import flair
import torch
from rapidfuzz import fuzz
from polyfuzz import PolyFuzz
from typing import List, Callable, Tuple
from bdikit.value_matching.base import BaseValueMatcher, ValueMatch
from polyfuzz.models import EditDistance, TFIDF, Embeddings
from flair.embeddings import TransformerWordEmbeddings, WordEmbeddings
from bdikit.config import get_device, VALUE_MATCHING_THRESHOLD


flair.device = torch.device(get_device())


class PolyFuzzValueMatcher(BaseValueMatcher):
    """
    Base class for value matching algorithms based on the PolyFuzz library.
    """

    def __init__(self, polyfuzz_model: PolyFuzz, threshold: float):
        self.model = polyfuzz_model
        self.threshold = threshold

    def match(
        self,
        source_values: List[str],
        target_values: List[str],
    ) -> List[ValueMatch]:

        self.model.match(source_values, target_values)
        match_results = self.model.get_matches()
        match_results.sort_values(by="Similarity", ascending=False, inplace=True)

        matches = []
        for _, row in match_results.iterrows():
            source = row[0]
            top_matches = row[1:]
            indexes = range(0, len(top_matches) - 1, 2)

            for index in indexes:
                target = top_matches[index]
                similarity = top_matches[index + 1]
                if similarity >= self.threshold:
                    matches.append(ValueMatch(source, target, similarity))

        return matches


class TFIDFValueMatcher(PolyFuzzValueMatcher):
    """
    Value matching algorithm based on the TF-IDF similarity between values.
    """

    def __init__(
        self,
        n_gram_range: Tuple[int, int] = (1, 3),
        clean_string: bool = True,
        threshold: float = VALUE_MATCHING_THRESHOLD,
        top_k: int = 1,
        cosine_method: str = "sparse",
    ):

        super().__init__(
            PolyFuzz(
                method=TFIDF(
                    n_gram_range=n_gram_range,
                    clean_string=clean_string,
                    min_similarity=threshold,
                    top_n=top_k,
                    cosine_method=cosine_method,
                )
            ),
            threshold,
        )


class EditDistanceValueMatcher(PolyFuzzValueMatcher):
    """
    Value matching algorithm based on the edit distance between values.
    """

    def __init__(
        self,
        scorer: Callable[[str, str], float] = fuzz.ratio,
        n_jobs: int = -1,
        threshold: float = VALUE_MATCHING_THRESHOLD,
    ):
        # Return scores between 0 and 1
        normalized_scorer = lambda str1, str2: scorer(str1, str2) / 100.0
        super().__init__(
            PolyFuzz(
                method=EditDistance(
                    n_jobs=n_jobs, scorer=normalized_scorer, normalize=False
                )
            ),
            threshold,
        )


class EmbeddingValueMatcher(PolyFuzzValueMatcher):
    """
    Value matching algorithm based on the cosine similarity of value embeddings.
    """

    def __init__(
        self,
        model_name: str = "bert-base-multilingual-cased",
        threshold: float = VALUE_MATCHING_THRESHOLD,
        top_k: int = 1,
        cosine_method: str = "sparse",
    ):
        embeddings = TransformerWordEmbeddings(model_name)
        method = Embeddings(
            embeddings,
            min_similarity=threshold,
            top_n=top_k,
            cosine_method=cosine_method,
        )
        super().__init__(PolyFuzz(method), threshold)


class FastTextValueMatcher(PolyFuzzValueMatcher):
    """
    Value matching algorithm based on the cosine similarity of FastText embeddings.
    """

    def __init__(
        self,
        model_name: str = "en-crawl",
        threshold: float = VALUE_MATCHING_THRESHOLD,
        top_k: int = 1,
        cosine_method: str = "sparse",
    ):
        embeddings = WordEmbeddings(model_name)
        method = Embeddings(
            embeddings,
            min_similarity=threshold,
            top_n=top_k,
            cosine_method=cosine_method,
        )
        super().__init__(PolyFuzz(method), threshold)



from abc import ABCMeta, abstractmethod
from typing import List
import pandas as pd
import numpy as np


class ColumnEmbedder(metaclass=ABCMeta):
    """
    Base class for column embedding algorithms. Implementations of this class
    must create embeddings for each of the columns of a table.
    """

    @abstractmethod
    def get_embeddings(self, table: pd.DataFrame) -> List[np.ndarray]:
        """
        Must compute a vector embedding for each column in the table.
        The vectors must be represented as dense np.ndarray vectors and
        appear in the same order they appear in the DataFrame `table`.
        """
        pass



import json
import pandas as pd
from os.path import join, dirname
from typing import List, Dict
from bdikit.standards.base import BaseStandard


GDC_SCHEMA_PATH = join(dirname(__file__), "../resource/gdc_schema.json")


class GDC(BaseStandard):
    """
    Class for GDC standard.
    """

    def __init__(self) -> None:
        self.data = None
        self.__read_data()

    def __read_data(self):
        with open(GDC_SCHEMA_PATH) as json_file:
            self.data = json.load(json_file)

    def get_columns(self) -> List[str]:
        return list(self.data.keys())

    def get_column_values(
        self, column_names: List[str]
    ) -> Dict[str, List]:  # get_gdc_data
        column_values = {}

        for column_name in column_names:
            raw_metadata = self.data.get(column_name, {})
            column_values[column_name] = list(raw_metadata.get("value_data", {}).keys())

        return column_values

    def get_column_metadata(
        self, column_names: List[str]
    ) -> Dict[str, Dict]:  # get_gdc_metadata
        column_metadata = {}

        for column_name in column_names:
            raw_metadata = self.data.get(column_name, {})
            column_metadata[column_name] = {}
            column_metadata[column_name]["description"] = raw_metadata.get(
                "column_description", ""
            )
            column_metadata[column_name]["value_names"] = list(
                raw_metadata.get("value_data", {}).keys()
            )
            column_metadata[column_name]["value_descriptions"] = list(
                raw_metadata.get("value_data", {}).values()
            )

        return column_metadata

    def get_dataframe_rep(self) -> pd.DataFrame:
        reshaped_data = {
            key: list(value["value_data"].keys()) for key, value in self.data.items()
        }

        # Ensure all lists have the same length by padding with None
        max_length = max(len(v) for v in reshaped_data.values())
        for k, v in reshaped_data.items():
            reshaped_data[k].extend([None] * (max_length - len(v)))

        df = pd.DataFrame.from_dict(reshaped_data, orient="columns")

        return df



import pandas as pd
from bdikit.schema_matching.one2one.valentine import (
    SimFloodSchemaMatcher,
    JaccardSchemaMatcher,
    DistributionBasedSchemaMatcher,
    ComaSchemaMatcher,
    CupidSchemaMatcher,
)
from bdikit.schema_matching.one2one.twophase import TwoPhaseSchemaMatcher
from bdikit.schema_matching.one2one.contrastivelearning import ContrastiveLearningSchemaMatcher


def test_basic_column_mapping_algorithms():
    for column_matcher in [
        SimFloodSchemaMatcher(),
        JaccardSchemaMatcher(),
        DistributionBasedSchemaMatcher(),
        ComaSchemaMatcher(),
        CupidSchemaMatcher(),
        #
        # Uncomment the following lines to test matchers that require
        # downloading large models
        #
        TwoPhaseSchemaMatcher(schema_matcher=ComaSchemaMatcher()),
        ContrastiveLearningSchemaMatcher(),
    ]:
        # given
        table1 = pd.DataFrame(
            {"column_1": ["a1", "b1", "c1"], "col_2": ["a2", "b2", "c2"]}
        )
        table2 = pd.DataFrame(
            {"column_1a": ["a1", "b1", "c1"], "col2": ["a2", "b2", "c2"]}
        )

        # when
        mapping = column_matcher.map(source=table1, target=table2)

        # then
        assert {
            "column_1": "column_1a",
            "col_2": "col2",
        } == mapping, f"{type(column_matcher).__name__} failed to map columns"



import collections
import math
import string

import pandas as pd
from pandas.api.types import infer_dtype


def computeTfIdf(tableDf):
    """Compute tfIdf of each column independently
        Called by _tokenize() method in dataset.py
    Args:
        table (DataFrame): input table
    Return: tfIdf dict containing tfIdf scores for all columns
    """

    # tfidf that considers each column (document) independently
    def computeTf(wordDict, doc):
        # input doc is a list
        tfDict = {}
        docCount = len(doc)
        for word, count in wordDict.items():
            tfDict[word] = count / float(docCount)
        return tfDict

    def computeIdf(docList):
        idfDict = {}
        N = len(docList)
        idfDict = dict.fromkeys(docList.keys(), 0)
        for word, val in docList.items():
            if val > 0:
                idfDict[word] += 1
        for word, val in idfDict.items():
            idfDict[word] = math.log10(N / float(val))
        return idfDict

    idf = {}
    for column in tableDf.columns:
        colVals = [val for entity in tableDf[column] for val in str(entity).split(" ")]
        wordSet = set(colVals)
        wordDict = dict.fromkeys(wordSet, 0)
        for val in colVals:
            wordDict[str(val)] += 1
        idf.update(computeIdf(wordDict))
    return idf


def pmiSample(val_counts, table, colIdxs, currIdx, max_tokens):
    """Compute PMI of pairs of columns (one of which is the topic column)
        Used in pmi sampling
    Args:
        val_counts (dict): stores the count of each (topic value, property value), topic value, and property value
        table (DataFrame): input table
        colIdxs (list): list of column indexes using column headers
        currIdx: current column index
        max_tokens: maximum tokens from pretrain arguments
    Return: list of sampled tokens for this column
    """
    tokens = []
    valPairs = []
    topicCol = table[colIdxs[0]]
    PMIs = {}
    for i in range(topicCol.shape[0]):
        topicVal = topicCol[i]
        propVal = table.at[i, currIdx]
        if (
            (topicVal, propVal) in val_counts
            and topicVal in val_counts
            and propVal in val_counts
        ):
            pair_pmi = val_counts[(topicVal, propVal)] / (
                val_counts[topicVal] * val_counts[propVal]
            )
            PMIs[(topicVal, propVal)] = pair_pmi
    PMIs = {
        k: v for k, v in sorted(PMIs.items(), key=lambda item: item[1], reverse=True)
    }
    if colIdxs.index(currIdx) == 0:
        valPairs = [k[0] for k in PMIs.keys()]
    else:
        valPairs = [k[1] for k in PMIs.keys()]
    for val in valPairs:
        for v in str(val).split(" "):
            if v not in tokens:
                tokens.append(v)
        if len(tokens) >= max_tokens:
            break
    return tokens


def constantSample(colVals, max_tokens):
    """Helper for preprocess() for constant sampling: take nth elements of column
        For sampling method 'constant'
    Args:
        colVals: list of tokens in each entity in the column
        max_tokens: maximum tokens specified in pretrain argument
    Return:
        list of tokens, such that list is of length max_tokens
    """
    step = math.ceil(len(colVals) / max_tokens)
    tokens = colVals[::step]
    while len(tokens) > max_tokens:
        step += 1
        tokens = colVals[::step]
    return tokens


def frequentSample(colVals, max_tokens):
    """Frequent sampling: Take most frequently occuring tokens
        For sampling method 'frequent'
    Args:
        colVals: list of tokens in each entity in the column
        max_tokens: maximum tokens specified in pretrain argument
    Return list of tokens
    """
    tokens, tokenFreq = [], {}
    tokenFreq = collections.Counter(colVals)
    tokenFreq = {
        k: v
        for k, v in sorted(tokenFreq.items(), key=lambda item: item[1], reverse=True)[
            :max_tokens
        ]
    }
    for t in colVals:
        if t in tokenFreq and t not in tokens:
            tokens.append(t)
    return tokens


def tfidfSample(column, tfidfDict, method, max_tokens):
    """TFIDF sampling: Take tokens with highest idf scores
        For sampling methods 'tfidf_token', 'tfidf_entity'
    Args:
        column (pandas Series): current column from input table DataFrame
        tfidfDict (dict): dict with tfidf scores for each column, created in _tokenize()
        method (str): sampling method ('tfidf_token', 'tfidf_entity')
        max_tokens: maximum tokens specified in pretrain argument
    Return list of tokens
    """
    tokens, tokenList, tokenFreq = [], [], {}
    if method == "tfidf_token":
        # token level
        for colVal in column.unique():
            for val in str(colVal).split(" "):
                idf = tfidfDict[val]
                tokenFreq[val] = idf
                tokenList.append(val)
        tokenFreq = {
            k: v
            for k, v in sorted(
                tokenFreq.items(), key=lambda item: item[1], reverse=True
            )[:max_tokens]
        }
        for t in tokenList:
            if t in tokenFreq and t not in tokens:
                tokens.append(t)

    elif method == "tfidf_entity":
        # entity level
        for colVal in column.unique():
            valIdfs = []
            for val in str(colVal).split(" "):
                valIdfs.append(tfidfDict[val])
            idf = sum(valIdfs) / len(valIdfs)
            tokenFreq[colVal] = idf
            tokenList.append(colVal)
        tokenFreq = {
            k: v
            for k, v in sorted(
                tokenFreq.items(), key=lambda item: item[1], reverse=True
            )
        }
        valCount, N = 0, 0
        for entity in tokenFreq:
            valCount += len(str(entity).split(" "))
            if valCount < max_tokens:
                N += 1
        tokenFreq = {k: tokenFreq[k] for k in list(tokenFreq)[:N]}
        for t in tokenList:
            if t in tokenFreq and t not in tokens:
                tokens += str(t).split(" ")
    return tokens


def tfidfRowSample(table, tfidfDict, max_tokens):
    """TFIDF sampling: Take rows with tokens that have highest idf scores
        For sampling method 'tfidf_row'
        Called in _tokenize() method in dataset.py
    Args:
        table (DataFrame): input table
        tfidfDict (dict): dict with tfidf scores for each column, created in _tokenize()
        max_tokens: maximum tokens specified in pretrain argument
    Return table with sampled rows using tfidf
    """
    tokenFreq = {}
    sortedRowInds = []
    for row in table.itertuples():
        index = row.Index
        valIdfs = []
        rowVals = [val for entity in list(row[1:]) for val in str(entity).split(" ")]
        for val in rowVals:
            valIdfs.append(tfidfDict[val])
        idf = sum(valIdfs) / len(valIdfs)
        tokenFreq[index] = idf
        tokenFreq = {
            k: v
            for k, v in sorted(
                tokenFreq.items(), key=lambda item: item[1], reverse=True
            )
        }
        sortedRowInds = list(tokenFreq.keys())[:max_tokens]
    table = table.reindex(sortedRowInds)
    return table


def preprocess(column: pd.Series, tfidfDict: dict, max_tokens: int, method: str):
    """Preprocess a column into list of max_tokens number of tokens
       Possible methods = "head", "alphaHead", "random", "constant", "frequent", "tfidf_token", "tfidf_entity", "tfidf_row"
    Args:
        column (pandas Series): current column from input table DataFrame
        tfidfDict (dict): dict with tfidf scores for each column, created in _tokenize()
        max_tokens: maximum tokens specified in pretrain argument
        method (str): sampling method from list of possible methods
    Returns list of sampled tokens
    """
    tokens = []
    colVals = [str(entity) for entity in column]
    if method == "head" or method == "tfidf_row":
        for val in colVals:
            if val not in tokens:
                tokens.append(val)
                if len(tokens) >= max_tokens:
                    break
    elif method == "alphaHead":
        if "mixed" in infer_dtype(column):
            column = column.astype(str)
        sortedCol = column.sort_values()
        sortedColVals = [
            str(val).lower() for entity in sortedCol for val in str(entity).split(" ")
        ]
        for val in sortedColVals:
            if val not in tokens:
                tokens.append(val)
                if len(tokens) >= max_tokens:
                    break
    elif method == "random":
        tokens = (
            pd.Series(colVals)
            .sample(min(len(colVals), max_tokens))
            .sort_index()
            .tolist()
        )
    elif method == "constant":
        tokens = constantSample(colVals, max_tokens)
    elif method == "frequent":
        tokens = frequentSample(colVals, max_tokens)
    elif "tfidf" in method and method != "tfidf_row":
        tokens = tfidfSample(column, tfidfDict, method, max_tokens)
    return tokens



import pandas as pd
from openai import OpenAI
from bdikit.schema_matching.one2one.base import BaseSchemaMatcher


class GPTSchemaMatcher(BaseSchemaMatcher):
    def __init__(self):
        self.client = OpenAI()

    def map(self, source: pd.DataFrame, target: pd.DataFrame):
        target_columns = target.columns
        labels = ", ".join(target_columns)
        candidate_columns = source.columns
        mappings = {}
        for column in candidate_columns:
            col = source[column]
            values = col.drop_duplicates().dropna()
            if len(values) > 15:
                rows = values.sample(15).tolist()
            else:
                rows = values.tolist()
            serialized_input = f"{column}: {', '.join([str(row) for row in rows])}"
            context = serialized_input.lower()
            column_types = self.get_column_type(context, labels)
            for column_type in column_types:
                if column_type in target_columns:
                    mappings[column] = column_type
                    break
        return self._fill_missing_matches(source, mappings)

    def get_column_type(
        self, context: str, labels: str, m: int = 10, model: str = "gpt-4-turbo-preview"
    ):
        messages = [
            {"role": "system", "content": "You are an assistant for column matching."},
            {
                "role": "user",
                "content": """ Please select the top """
                + str(m)
                + """ class from """
                + labels
                + """ which best describes the context. The context is defined by the column name followed by its respective values. Please respond only with the name of the classes separated by semicolon.
                    \n CONTEXT: """
                + context
                + """ \n RESPONSE: \n""",
            },
        ]
        col_type = self.client.chat.completions.create(
            model=model, messages=messages, temperature=0.3
        )
        col_type_content = col_type.choices[0].message.content
        return col_type_content.split(";")



import os
import torch

os.environ["TOKENIZERS_PARALLELISM"] = "false"


BDIKIT_DEVICE: str = os.getenv("BDIKIT_DEVICE", default="cpu")
VALUE_MATCHING_THRESHOLD = 0.3
DEFAULT_VALUE_MATCHING_METHOD = "tfidf"
DEFAULT_SCHEMA_MATCHING_METHOD = "ct_learning"


def get_device() -> str:
    if BDIKIT_DEVICE == "auto":
        return "cuda" if torch.cuda.is_available() else "cpu"
    else:
        return BDIKIT_DEVICE



import pandas as pd
import numpy as np
from bdikit.mapping_functions import (
    FunctionValueMapper,
    DictionaryMapper,
    IdentityValueMapper,
)


def test_identity_mapper():
    # given
    str_column = pd.Series(data=["a", "b", "c", "d", "e"], name="column_str")
    identity_mapper = IdentityValueMapper()

    # when
    mapped_column = identity_mapper.map(str_column)

    # then
    assert mapped_column.eq(["a", "b", "c", "d", "e"]).all()


def test_dictionary_mapper():
    # given
    str_column = pd.Series(data=["a", "b", "c", "d", "e"], name="column_str")
    dict_mapper = DictionaryMapper(dictionary={"a": 1, "b": 2, "c": 3, "d": 4, "e": 5})

    # when
    mapped_column = dict_mapper.map(str_column)

    # then
    assert mapped_column.eq([1, 2, 3, 4, 5]).all()


def test_dictionary_mapper_missing_key():
    # given
    str_column = pd.Series(
        data=["a", "b", "c", "d", "e", None, np.nan], name="column_str"
    )

    # when
    dict_mapper = DictionaryMapper(
        dictionary={"a": "1", "b": "2", "c": "3", "d": "4"}, missing_key_value=np.nan
    )
    mapped_column = dict_mapper.map(str_column)
    # then
    assert mapped_column[0:4].tolist() == ["1", "2", "3", "4"]
    assert np.isnan(mapped_column[4])
    assert np.isnan(mapped_column[5])
    assert np.isnan(mapped_column[6])

    # when
    dict_mapper = DictionaryMapper(
        dictionary={"a": 1, "b": 2, "c": 3, "d": 4}, missing_key_value=np.nan
    )
    mapped_column = dict_mapper.map(str_column)
    # then
    assert mapped_column[0:4].tolist() == [1.0, 2.0, 3.0, 4.0]
    assert np.isnan(mapped_column[4])
    assert np.isnan(mapped_column[5])
    assert np.isnan(mapped_column[6])


def test_custom_function_mapper():
    # given
    str_column = pd.Series(data=["a", "b", "c", "d", "e"], name="column_str")
    fn_mapper = FunctionValueMapper(function=lambda x: x + x)

    # when
    mapped_column = fn_mapper.map(str_column)

    # then
    assert mapped_column.eq(["aa", "bb", "cc", "dd", "ee"]).all()



from typing import List, NamedTuple, TypedDict, Set


class ValueMatch(NamedTuple):
    """
    Represents a match between a source value and a target value with a
    similarity score.
    """

    source_value: str
    target_value: str
    similarity: float


class ValueMatchingResult(TypedDict):
    """
    Represents the result of a value matching operation.
    """

    source: str
    target: str
    matches: List[ValueMatch]
    coverage: float
    unique_values: Set[str]
    unmatch_values: Set[str]


class BaseValueMatcher:
    """
    Base class for value matching algorithms, i.e., algorithms that match
    values from a source domain to values from a target domain.
    """

    def match(
        self, source_values: List[str], target_values: List[str]
    ) -> List[ValueMatch]:
        raise NotImplementedError("Subclasses must implement this method")



import pandas as pd
import torch
from bdikit.models.contrastive_learning.cl_preprocessor import (
    preprocess,
)
from torch.utils import data
from transformers import AutoTokenizer

lm_mp = {
    "roberta": "roberta-base",
    "bert": "bert-base-uncased",
    "distilbert": "distilbert-base-uncased",
}


class PretrainTableDataset(data.Dataset):
    def __init__(self, max_len=128, lm="roberta", sample_meth="head"):
        super().__init__()
        self.max_len = max_len
        self.tokenizer = AutoTokenizer.from_pretrained(lm_mp[lm])
        self.sample_meth = sample_meth

    def _tokenize(self, table: pd.DataFrame):
        res = []
        max_tokens = self.max_len * 2 // len(table.columns)
        budget = max(1, self.max_len // len(table.columns) - 1)
        tfidfDict = None

        column_mp = {}

        for column in table.columns:
            tokens = preprocess(
                table[column], tfidfDict, max_tokens, self.sample_meth
            )  # from preprocessor.py
            col_text = (
                self.tokenizer.cls_token
                + column
                + self.tokenizer.sep_token
                + self.tokenizer.sep_token.join(tokens[:max_tokens])
            )
            column_mp[column] = len(res)
            res += self.tokenizer.encode(
                text=col_text,
                max_length=budget,
                add_special_tokens=False,
                truncation=True,
            )
        return res, column_mp

    def pad(self, batch):
        """Merge a list of dataset items into a training batch

        Args:
            batch (list of tuple): a list of sequences

        Returns:
            LongTensor: x_ori of shape (batch_size, seq_len)
            LongTensor: x_aug of shape (batch_size, seq_len)
            tuple of List: the cls indices
        """
        x_ori, x_aug, cls_indices = zip(*batch)
        max_len_ori = max([len(x) for x in x_ori])
        max_len_aug = max([len(x) for x in x_aug])
        maxlen = max(max_len_ori, max_len_aug)
        x_ori_new = [
            xi + [self.tokenizer.pad_token_id] * (maxlen - len(xi)) for xi in x_ori
        ]
        x_aug_new = [
            xi + [self.tokenizer.pad_token_id] * (maxlen - len(xi)) for xi in x_aug
        ]

        # decompose the column alignment
        cls_ori = []
        cls_aug = []
        for item in cls_indices:
            cls_ori.append([])
            cls_aug.append([])

            for idx1, idx2 in item:
                cls_ori[-1].append(idx1)
                cls_aug[-1].append(idx2)

        return (
            torch.LongTensor(x_ori_new),
            torch.LongTensor(x_aug_new),
            (cls_ori, cls_aug),
        )



import importlib
from enum import Enum
from typing import Mapping, Any
from bdikit.standards.base import BaseStandard


class Standards(Enum):
    GDC = ("gdc", "bdikit.standards.gdc.GDC")

    def __init__(self, standard_name: str, standard_path: str):
        self.standard_name = standard_name
        self.standard_path = standard_path

    @staticmethod
    def get_standard(
        standard_name: str, **standard_kwargs: Mapping[str, Any]
    ) -> BaseStandard:
        if standard_name not in standards:
            names = ", ".join(list(standards.keys()))
            raise ValueError(
                f"The {standard_name} standard is not supported. "
                f"Supported standards are: {names}"
            )
        # Load the class dynamically
        module_path, class_name = standards[standard_name].rsplit(".", 1)
        module = importlib.import_module(module_path)

        return getattr(module, class_name)(**standard_kwargs)


standards = {standard.standard_name: standard.standard_path for standard in Standards}



import os
import setuptools


package_name = 'bdi-kit'
package_dir = 'bdikit'


def read_readme():
    with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf8') as file:
        return file.read()


def read_version():
    module_path = os.path.join(package_dir, '__init__.py')
    with open(module_path) as file:
        for line in file:
            parts = line.strip().split(' ')
            if parts and parts[0] == '__version__':
                return parts[-1].strip("'").strip("\"")

    raise KeyError('Version not found in {0}'.format(module_path))


def get_requires():
    with open('requirements.txt') as fp:
        dependencies = [line for line in fp if line and not line.startswith('#')]

        return dependencies


long_description = read_readme()
version = read_version()
requires = get_requires()
extra_requires = {}

setuptools.setup(
    name=package_name,
    version=version,
    packages=setuptools.find_packages(),
    install_requires=requires,
    extras_require=extra_requires,
    python_requires='>=3.9',
    description="bdi-kit library",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/VIDA-NYU/bdi-kit',
    include_package_data=True,
    author='',
    author_email='',
    maintainer='',
    maintainer_email='',
    keywords=['askem', 'data integration', 'nyu'],
    license='Apache-2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering',
    ])


from abc import ABCMeta, abstractmethod
from typing import List, NamedTuple, TypedDict
import pandas as pd


class ColumnScore(NamedTuple):
    column_name: str
    score: float


class TopkMatching(TypedDict):
    source_column: str
    top_k_columns: List[ColumnScore]


class BaseTopkSchemaMatcher(metaclass=ABCMeta):
    @abstractmethod
    def get_recommendations(
        self, source: pd.DataFrame, target: pd.DataFrame, top_k: int
    ) -> List[TopkMatching]:
        pass
